Package: rdomains
Title: Get the Category of Content Hosted by a Domain
Version: 0.4.0
Authors@R: person("Gaurav", "Sood", email = "gsood07@gmail.com", role = c("aut", "cre"))
Description: Get the category of content hosted by a domain. Use Shallalist (service discontinued),
    'VirusTotal' (which provides access to lots of services) <https://www.virustotal.com/>,
    'DMOZ' <https://archive.org/details/dmoz-rdf-20150327>, University Domain list
    <https://github.com/Hipo/university-domains-list>, 'OpenAI' 'GPT' models, 'Anthropic' 'Claude' models,
    or validated machine learning classifiers based on 'Shallalist' data to learn about the kind
    of content hosted by a domain.
Depends: R (>= 4.1.0)
Imports: Matrix, urltools, glmnet, stats, methods, XML, httr, xml2,
        curl, virustotal, jsonlite, R.utils, dplyr (>= 1.1.0), purrr
        (>= 1.0.0), tibble (>= 3.2.0), stringr (>= 1.5.0), rlang (>=
        1.1.0), cli (>= 3.6.0), checkmate (>= 2.3.0), glue (>= 1.6.0),
        readr (>= 2.1.0)
Suggests: testthat, rmarkdown, knitr (>= 1.11)
VignetteBuilder: knitr
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-13 04:32:04 UTC; soodoku
Author: Gaurav Sood [aut, cre]
Maintainer: Gaurav Sood <gsood07@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-13 04:50:02 UTC
Built: R 4.6.0; ; 2026-05-13 05:51:52 UTC; unix

Package: rdomains
Title: Get the Category of Content Hosted by a Domain
Version: 0.5.0
Authors@R: person("Gaurav", "Sood", email = "gsood07@gmail.com", role = c("aut", "cre"))
Description: Get the category of content hosted by a domain. Use Shallalist (service discontinued),
    'VirusTotal' (which provides access to lots of services) <https://www.virustotal.com/>,
    'DMOZ' <https://archive.org/details/dmoz-rdf-20150327>, University Domain list
    <https://github.com/Hipo/university-domains-list>, 'OpenAI' 'GPT' models, 'Anthropic' 'Claude' models,
    or validated machine learning classifiers based on 'Shallalist' data to learn about the kind
    of content hosted by a domain.
Depends: R (>= 4.1.0)
Imports: Matrix, urltools, glmnet, stats, methods, XML, httr, httr2,
        xml2, curl, virustotal, jsonlite, R.utils, dplyr (>= 1.1.0),
        purrr (>= 1.0.0), tibble (>= 3.2.0), stringr (>= 1.5.0), rlang
        (>= 1.1.0), cli (>= 3.6.0), checkmate (>= 2.3.0), glue (>=
        1.6.0), readr (>= 2.1.0)
Suggests: testthat, rmarkdown, knitr (>= 1.11), withr
VignetteBuilder: knitr
License: MIT + file LICENSE
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-08-02 08:22:57 UTC; soodoku
Author: Gaurav Sood [aut, cre]
Maintainer: Gaurav Sood <gsood07@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-02 09:00:02 UTC

# Generated by roxygen2: do not edit by hand

export(adult_ml1_cat)
export(claude_cat)
export(dmoz_cat)
export(get_dmoz_data)
export(get_shalla_data)
export(get_stevenblack_data)
export(glm_shalla)
export(not_news)
export(openai_cat)
export(shalla_cat)
export(stevenblack_cat)
export(uni_cat)
export(virustotal_cat)
importFrom(Matrix,Matrix)
importFrom(Matrix,spMatrix)
importFrom(Matrix,sparseVector)
importFrom(R.utils,gunzip)
importFrom(XML,readHTMLTable)
importFrom(checkmate,assert_character)
importFrom(checkmate,assert_file_exists)
importFrom(checkmate,assert_logical)
importFrom(cli,cli_abort)
importFrom(cli,cli_inform)
importFrom(cli,cli_warn)
importFrom(curl,curl_download)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,coalesce)
importFrom(dplyr,mutate)
importFrom(glmnet,predict.glmnet)
importFrom(glue,glue)
importFrom(httr,GET)
importFrom(httr,POST)
importFrom(httr,add_headers)
importFrom(httr,content)
importFrom(httr,oauth_app)
importFrom(httr,oauth_signature)
importFrom(httr,status_code)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(methods,as)
importFrom(purrr,map_chr)
importFrom(purrr,map_df)
importFrom(readr,read_csv)
importFrom(rlang,.data)
importFrom(rlang,abort)
importFrom(rlang,inform)
importFrom(rlang,warn)
importFrom(stats,predict)
importFrom(stats,setNames)
importFrom(stringr,str_detect)
importFrom(stringr,str_remove)
importFrom(stringr,str_trim)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(urltools,suffix_extract)
importFrom(utils,URLencode)
importFrom(utils,read.csv)
importFrom(utils,read.table)
importFrom(utils,untar)
importFrom(utils,unzip)
importFrom(utils,write.csv)
importFrom(virustotal,domain_report)
importFrom(virustotal,set_key)
importFrom(xml2,as_list)
importFrom(xml2,read_xml)

## rdomains: Classify Domains Based on Their Content

[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/rdomains)](https://cran.r-project.org/package=rdomains) 
[![Build Status](https://travis-ci.org/themains/rdomains.svg?branch=master)](https://travis-ci.org/themains/rdomains)
[![Build status](https://ci.appveyor.com/api/projects/status/3vjmwn7jyf1s17e4?svg=true)](https://ci.appveyor.com/project/soodoku/rdomains)
[![codecov](https://codecov.io/github/themains/rdomains/branch/master/graph/badge.svg)](https://codecov.io/github/themains/rdomains/)
![](http://cranlogs.r-pkg.org/badges/grand-total/rdomains)

The package provides a few ways to classify domains based on their content. You can either get the categorizations from [shallalist](http://www.shallalist.de/) (which has stopped its service --- the latest you will get is from 1/14/22), [trusted (McAfee)](https://trustedsource.org/), DMOZ (the service has ended; available at [curlie](https://curlie.org/)), [Alexa API](https://docs.aws.amazon.com/AlexaWebInfoService/latest/), which uses the DMOZ Data (now hosted at [https://curlie.org](https://curlie.org)), or [virustotal API](http://virustotal.com), or use validated machine learning models based off the shallalist data.

### Installation

To get the current release version from CRAN:

```r
install.packages("rdomains")
```

To get the current development version from GitHub:

```r
# install.packages("devtools")
devtools::install_github("themains/rdomains", build_vignettes = TRUE)
```

### Usage

To learn how to use rdomains, launch the vignette within R:

```r
vignette("rdomains", package = "rdomains")
```

### License

Scripts are released under the [MIT License](https://opensource.org/licenses/MIT).

#' Get Steven Black's Host List Data
#'
#' Downloads the latest version of Steven Black's unified hosts file.
#' The hosts file contains domains known for serving ads, malware, and tracking.
#'
#' @param outdir    Optional; folder to which you want to save the file; Default is current directory
#' @param variant   Optional; which variant to download. Options: "base", "porn", "social", "gambling", "all"
#' @param overwrite Optional; default is FALSE. If TRUE, the file is overwritten.
#' 
#' @export
#' 
#' @references \url{https://github.com/StevenBlack/hosts}
#' 
#' @examples \dontrun{
#' get_stevenblack_data()
#' get_stevenblack_data(variant = "all")
#' }

get_stevenblack_data <- function(outdir = "./", variant = "base", overwrite = FALSE) {
  
  # Define available variants and their URLs
  variants <- list(
    base = "https://raw.githubusercontent.com/StevenBlack/hosts/master/hosts",
    porn = "https://raw.githubusercontent.com/StevenBlack/hosts/master/alternates/porn/hosts",
    social = "https://raw.githubusercontent.com/StevenBlack/hosts/master/alternates/social/hosts", 
    gambling = "https://raw.githubusercontent.com/StevenBlack/hosts/master/alternates/gambling/hosts",
    all = "https://raw.githubusercontent.com/StevenBlack/hosts/master/alternates/porn-social-gambling/hosts"
  )
  
  if (!variant %in% names(variants)) {
    stop("Invalid variant. Choose from: ", paste(names(variants), collapse = ", "))
  }
  
  # Create output filename
  output_file <- file.path(outdir, paste0("stevenblack_hosts_", variant, ".txt"))
  
  # Check if file already exists
  if (!overwrite && file.exists(output_file)) {
    stop("File already exists: ", output_file, 
         "\nSet overwrite=TRUE to replace it.")
  }
  
  # Create output directory if it doesn't exist
  if (!dir.exists(outdir)) {
    dir.create(outdir, recursive = TRUE)
  }
  
  # Download the hosts file
  tryCatch({
    cat("Downloading Steven Black's hosts file (", variant, " variant)...\n")
    curl::curl_download(variants[[variant]], output_file)
    cat("Steven Black's hosts data saved to:", output_file, "\n")
    
    # Print some statistics
    hosts_lines <- readLines(output_file, warn = FALSE)
    blocked_count <- sum(grepl("^(0\\.0\\.0\\.0|127\\.0\\.0\\.1)\\s+", hosts_lines))
    cat("Total blocked domains:", blocked_count, "\n")
    
  }, error = function(e) {
    stop("Failed to download hosts file: ", e$message)
  })
  
  invisible(output_file)
}

test_that("stevenblack_cat returns expected data structure", {
  # Test with minimal mock data
  mock_hosts <- c(
    "# This is a test hosts file",
    "127.0.0.1 localhost",
    "0.0.0.0 doubleclick.net",
    "0.0.0.0 googleadservices.com",
    "127.0.0.1 malware-example.com"
  )
  
  # Create temporary hosts file
  temp_hosts <- tempfile(fileext = ".txt")
  writeLines(mock_hosts, temp_hosts)
  
  # Test function
  result <- stevenblack_cat(c("google.com", "doubleclick.net"), use_file = temp_hosts)
  
  # Clean up
  unlink(temp_hosts)
  
  # Check structure. The third column is the vintage of the list that produced the
  # category -- see source_vintage(). For this list it is the file's own date, because
  # Steven Black's hosts file is actively maintained.
  expect_s3_class(result, "data.frame")
  expect_equal(ncol(result), 3)
  expect_equal(names(result), c("domain", "stevenblack", "source_last_published"))
  expect_equal(nrow(result), 2)
  
  # Check classifications
  expect_equal(result$stevenblack[1], "safe")  # google.com not in blocklist
  expect_equal(result$stevenblack[2], "ads")   # doubleclick.net classified as ads
})

test_that("stevenblack_cat handles domain cleaning", {
  mock_hosts <- c("0.0.0.0 example.com")
  temp_hosts <- tempfile(fileext = ".txt")
  writeLines(mock_hosts, temp_hosts)
  
  # Test with various domain formats
  domains <- c(
    "http://example.com",
    "https://www.example.com",
    "example.com/path"
  )
  
  result <- stevenblack_cat(domains, use_file = temp_hosts)
  unlink(temp_hosts)
  
  # All should be classified as blocked since they resolve to example.com
  expect_equal(sum(result$stevenblack == "blocked"), 3)
})

test_that("stevenblack_cat handles errors appropriately", {
  # Test with non-existent file
  expect_error(
    stevenblack_cat("test.com", use_file = "nonexistent.txt"),
    "File does not exist"
  )

  # Test with no domain
  expect_error(
    stevenblack_cat(NULL),
    "must not be NULL"
  )
})

test_that("get_stevenblack_data creates file", {
  skip_if_offline()
  
  temp_dir <- tempdir()
  
  # Test download (this might be slow, so we'll skip on CRAN)
  skip_on_cran()
  
  # Download base variant
  result_file <- get_stevenblack_data(outdir = temp_dir, variant = "base", overwrite = TRUE)
  
  expect_true(file.exists(result_file))
  expect_true(file.size(result_file) > 1000)  # Should be a substantial file
  
  # Clean up
  unlink(result_file)
})

test_that("get_stevenblack_data validates variants", {
  expect_error(
    get_stevenblack_data(variant = "invalid"),
    "Invalid variant"
  )
})

test_that("get_stevenblack_data respects overwrite parameter", {
  temp_dir <- tempdir()
  test_file <- file.path(temp_dir, "stevenblack_hosts_base.txt")
  
  # Create a dummy file
  writeLines("test", test_file)
  
  # Should error without overwrite
  expect_error(
    get_stevenblack_data(outdir = temp_dir, overwrite = FALSE),
    "File already exists"
  )
  
  # Clean up
  unlink(test_file)
})

#' Classify Domain Using OpenAI (Internal Helper)
#'
#' @param domain single domain name
#' @param categories allowed categories
#' @param model OpenAI model name
#' @param api_key API key
#' @keywords internal
#' @noRd
classify_domain_openai <- function(domain, categories, model, api_key) {
  prompt <- build_categorization_prompt(domain, categories)

  request_body <- list(
    model = model,
    messages = list(
      list(role = "user", content = prompt)
    ),
    max_tokens = 50,
    temperature = 0.1
  )

  result <- tryCatch({
    response <- httr::POST(
      url = "https://api.openai.com/v1/chat/completions",
      httr::add_headers(
        "Authorization" = paste("Bearer", api_key),
        "Content-Type" = "application/json"
      ),
      body = jsonlite::toJSON(request_body, auto_unbox = TRUE),
      encode = "raw"
    )

    if (status_code(response) == 200) {
      result <- fromJSON(content(response, "text", encoding = "UTF-8"))
      category <- str_trim(result$choices[1, ]$message$content)

      if (category %in% categories) {
        category
      } else {
        "other"
      }
    } else {
      cli_warn("API call failed for domain {domain} - Status: {status_code(response)}")
      NA_character_
    }
  }, error = function(e) {
    cli_warn(c(
      "Error processing domain: {domain}",
      "x" = e$message
    ))
    NA_character_
  })

  tibble(domain_name = domain, openai_category = result)
}

#' Get Category from OpenAI
#'
#' Fetches category of content hosted by a domain using OpenAI's chat completion API.
#' The function uses GPT models to classify domains into specified categories.
#'
#' @param domains vector of domain names
#' @param api_key OpenAI API key. If not provided, looks for OPENAI_API_KEY environment variable
#' @param categories vector of categories to classify into. If NULL, uses default web categories
#' @param model OpenAI model to use (default: "gpt-4o-mini" for cost efficiency)
#' @param rate_limit delay in seconds between API calls (default: 0.5)
#'
#' @return data.frame with original list and content category of the domain
#'
#' @export
#' @examples \dontrun{
#' openai_cat("google.com")
#' openai_cat(c("google.com", "facebook.com"))
#' openai_cat("google.com", categories = c("search", "social", "ecommerce", "news", "other"))
#' }

openai_cat <- function(domains = NULL, api_key = NULL, categories = NULL,
                      model = "gpt-4o-mini", rate_limit = 0.5) {

  validate_domains(domains)
  c_domains <- clean_domains(domains)

  if (is.null(categories)) {
    categories <- c("news", "shopping", "social", "adult", "gambling", "technology",
                   "finance", "education", "entertainment", "business", "other")
  }

  api_key <- get_api_key(api_key, "OPENAI_API_KEY", "OpenAI")

  map_df(
    seq_along(c_domains),
    function(i) {
      result <- classify_domain_openai(c_domains[i], categories, model, api_key)
      apply_rate_limit(i, length(c_domains), rate_limit)
      result
    }
  )
}

test_that("openai cat validates input", {
  expect_error(openai_cat(), "must not be NULL")

  # Temporarily unset API key to test error handling
  old_key <- Sys.getenv("OPENAI_API_KEY")
  Sys.unsetenv("OPENAI_API_KEY")
  expect_error(openai_cat("google.com"), "API key not found")
  # Restore the key
  if (old_key != "") Sys.setenv(OPENAI_API_KEY = old_key)
})

test_that("openai cat returns correct structure", {
  skip_if(identical(Sys.getenv("OPENAI_API_KEY"), ""), "OPENAI_API_KEY not set")
  
  result <- openai_cat("google.com")
  expect_s3_class(result, "data.frame")
  expect_named(result, c("domain_name", "openai_category"))
  expect_equal(nrow(result), 1)
  expect_equal(result$domain_name[1], "google.com")
})

test_that("openai cat handles multiple domains", {
  skip_if(identical(Sys.getenv("OPENAI_API_KEY"), ""), "OPENAI_API_KEY not set")
  
  domains <- c("google.com", "facebook.com")
  result <- openai_cat(domains)
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_equal(result$domain_name, domains)
})

test_that("openai cat preprocesses domains correctly", {
  skip_if(identical(Sys.getenv("OPENAI_API_KEY"), ""), "OPENAI_API_KEY not set")
  
  result <- openai_cat("http://www.google.com")
  expect_equal(result$domain_name[1], "google.com")
})

test_that("openai cat accepts custom categories", {
  skip_if(identical(Sys.getenv("OPENAI_API_KEY"), ""), "OPENAI_API_KEY not set")
  
  custom_cats <- c("search", "social", "other")
  result <- openai_cat("google.com", categories = custom_cats)
  expect_s3_class(result, "data.frame")
  # Category should be one of the custom categories or NA
  expect_true(is.na(result$openai_category[1]) || result$openai_category[1] %in% custom_cats)
})

#' Get Shalla Data
#'
#' Shallalist service was discontinued in January 2022. This function downloads 
#' the last archived copy (from 1/14/22) that we have preserved on GitHub.
#' The original service at shallalist.de is no longer available.
#' Downloads, unzips and saves the final version of shallalist data. By default, saves shalla data  
#' as \code{shalla_domain_category.csv}.
#'  
#' @param outdir    Optional; folder to which you want to save the file; Default is same folder
#' @param overwrite Optional; default is FALSE. If TRUE, the file is overwritten.
#' 
#' @export
#' 
#' @references \url{https://web.archive.org/web/20210502020725/http://www.shallalist.de/}
#' 
#' @examples \dontrun{
#' get_shalla_data()
#' }

get_shalla_data <- function(outdir = "./", overwrite = FALSE) {

  # Normalize and create output directory path
  outdir <- normalizePath(outdir, mustWork = FALSE)
  if (!dir.exists(outdir)) {
    tryCatch({
      dir.create(outdir, recursive = TRUE)
    }, error = function(e) {
      stop("Cannot create output directory: ", outdir, "\n", 
           "Error: ", e$message, "\n",
           "Please check directory permissions.")
    })
  }
  
  # Use proper file path construction
  output_file <- file.path(outdir, "shalla_domain_category.csv")
  
  # Check if file already exists
  if (!overwrite && file.exists(output_file)) {
    stop("File already exists: ", output_file, "\n",
         "Set overwrite=TRUE to replace it or choose a different location.")
  }

  # Create temporary file in a writable location
  tmp_dir <- tempdir()
  tmp <- tempfile(tmpdir = tmp_dir, fileext = ".gz")
  
  tryCatch({
    # Download file
    cat("Downloading Shallalist data...\n")
    curl::curl_download(
      "https://raw.githubusercontent.com/themains/rdomains/master/data-raw/shallalist/accomplist/shallalist.gz", 
      tmp
    )
    
    # Extract to destination with proper error handling
    R.utils::gunzip(tmp, destname = output_file, overwrite = overwrite)
    
    # Verify the file was created successfully
    if (!file.exists(output_file)) {
      stop("Failed to create output file. Please check write permissions for: ", outdir)
    }
    
    cat("Shallalist data saved to:", output_file, "\n")
    
  }, error = function(e) {
    # Clean up temp file on error
    if (file.exists(tmp)) unlink(tmp, force = TRUE)
    
    if (grepl("permission", e$message, ignore.case = TRUE)) {
      stop("Permission denied. Please ensure you have write access to: ", outdir, "\n",
           "On Windows, try running R as administrator or choose a different output directory.")
    } else {
      stop("Error downloading or extracting Shallalist data: ", e$message)
    }
  })
  
  # Clean up temp file
  if (file.exists(tmp)) unlink(tmp, force = TRUE)
  
  invisible(output_file)
}

test_that("shallalist cat", {
  # Downloads a real file: CRAN forbids that, and it fails offline.
  skip_on_cran()
  skip_if_offline()

  # Use temporary directory to avoid conflicts
  temp_dir <- tempdir()
  get_shalla_data(outdir = temp_dir, overwrite = TRUE)

  # Test with the downloaded data
  data_file <- file.path(temp_dir, "shalla_domain_category.csv")
  reset_vintage_warnings()
  expect_warning(
    report <- shalla_cat("http://www.google.com", use_file = data_file),
    "no longer published"
  )

  expect_s3_class(report, "data.frame")
  expect_equal(report$source_last_published, "2022-01")

  # Clean up
  if (file.exists(data_file)) {
    unlink(data_file)
  }
})

test_that("shalla_cat validates input", {
  expect_error(shalla_cat(NULL), "must not be NULL")
  expect_error(shalla_cat(""), "contains empty strings")
  expect_error(shalla_cat(123), "character")
})

test_that("shalla_cat provides helpful error for missing file", {
  expect_error(
    shalla_cat("google.com", use_file = "nonexistent.csv"),
    "File not found"
  )
})

#' Get Category from Steven Black's Host List
#'
#' Classifies domains based on Steven Black's unified host list which blocks
#' ads, malware, and tracking domains. The function checks if a domain appears
#' in the blocklist and categorizes it accordingly.
#'
#' Steven Black's host list is a consolidated list from multiple sources including
#' adaway.org, mvps.org, malwaredomainlist.com, and someonewhocares.org.
#'
#' Unlike the Shallalist and DMOZ lookups, this list is actively maintained, so the
#' returned \code{source_last_published} is the fetched file's own modification date
#' rather than a constant. See \code{\link{source_vintage}}.
#'
#' @param domain domain names as character vector
#' @param use_file path to a local Steven Black hosts file. If NULL, downloads from GitHub
#'   once per session and reuses it
#'
#' @return data.frame with the original domain name, the category, and the date of the
#'   list that supplied it
#'
#' @export
#' @seealso \code{\link{source_vintage}} for the provenance of every category source
#' @references \url{https://github.com/StevenBlack/hosts}
#'
#' @examples \dontrun{
#' stevenblack_cat("doubleclick.net")
#' stevenblack_cat(c("google.com", "googleadservices.com", "malware-example.com"))
#' }

stevenblack_cat <- function(domain = NULL, use_file = NULL) {

  validate_domains(domain, "domain")
  clean_doms <- clean_domains(domain)

  downloaded <- FALSE
  if (is.null(use_file)) {
    # Cached for the session. This used to re-download ~4 MB on *every* call, so
    # classifying domains one at a time cost a fresh download each time.
    cached <- .rdomains_env$stevenblack_file
    if (!is.null(cached) && file.exists(cached)) {
      hosts_file <- cached
    } else {
      hosts_file <- tempfile(fileext = ".hosts")
      tryCatch({
        cli_inform("Downloading Steven Black's hosts file...")
        curl::curl_download(
          "https://raw.githubusercontent.com/StevenBlack/hosts/master/hosts",
          hosts_file
        )
      }, error = function(e) {
        cli_abort(c(
          "Failed to download hosts file",
          "x" = e$message
        ))
      })
      .rdomains_env$stevenblack_file <- hosts_file
      downloaded <- TRUE
    }
  } else {
    assert_file_exists(use_file)
    hosts_file <- use_file
  }

  file_date <- tryCatch(
    format(as.Date(file.info(hosts_file)$mtime), "%Y-%m-%d"),
    error = function(e) NA_character_
  )

  hosts_lines <- tryCatch({
    readLines(hosts_file, warn = FALSE)
  }, error = function(e) {
    cli_abort(c(
      "Failed to read hosts file",
      "x" = e$message
    ))
  })

  blocked_pattern <- "^(0\\.0\\.0\\.0|127\\.0\\.0\\.1)\\s+"
  blocked_lines <- hosts_lines[str_detect(hosts_lines, blocked_pattern)]

  blocked_domains <- blocked_lines |>
    str_remove(blocked_pattern) |>
    str_trim() |>
    (\(x) x[!str_detect(x, "^#")])() |>
    (\(x) x[x != ""])() |>
    (\(x) x[x != "localhost"])()

  results <- map_df(seq_along(clean_doms), function(i) {
    category <- if (clean_doms[i] %in% blocked_domains) {
      if (str_detect(clean_doms[i], stringr::regex("ad|ads|doubleclick|googleadservices|googlesyndication", ignore_case = TRUE))) {
        "ads"
      } else if (str_detect(clean_doms[i], stringr::regex("malware|virus|trojan|phishing", ignore_case = TRUE))) {
        "malware"
      } else if (str_detect(clean_doms[i], stringr::regex("track|analytics|metric|stats", ignore_case = TRUE))) {
        "tracking"
      } else {
        "blocked"
      }
    } else {
      "safe"
    }
    tibble(
      domain = domain[i],
      stevenblack = category,
      source_last_published = file_date
    )
  })

  # The cached file is deliberately not unlinked -- it is the session cache. R removes the
  # tempdir on exit.
  invisible(downloaded)

  results
}

#' Clean and Normalize Domain Names
#'
#' @param domains character vector of domain names
#' @return character vector of cleaned domains
#' @keywords internal
#' @noRd
clean_domains <- function(domains) {
  domains |>
    str_trim() |>
    str_remove("^https?://") |>
    str_remove("^www\\.") |>
    str_remove("/.*$")
}

#' Validate Domain Input
#'
#' @param domains domain input to validate
#' @param arg_name name of argument for error messages
#' @keywords internal
#' @noRd
validate_domains <- function(domains, arg_name = "domains") {
  if (is.null(domains)) {
    cli_abort("{.arg {arg_name}} must not be NULL")
  }

  assert_character(
    domains,
    min.len = 1,
    any.missing = FALSE,
    .var.name = arg_name
  )

  if (any(str_trim(domains) == "")) {
    cli_abort("{.arg {arg_name}} contains empty strings")
  }

  invisible(domains)
}

#' Validate Data File Path
#'
#' @param file_path path to file
#' @param default_name default filename to look for
#' @param download_function name of function to download data
#' @keywords internal
#' @noRd
validate_data_file <- function(file_path = NULL,
                               default_name,
                               download_function) {
  if (!is.null(file_path)) {
    assert_character(file_path, len = 1)
    if (!file.exists(file_path)) {
      cli_abort(c(
        "File not found: {.file {file_path}}",
        "i" = "Download data using {.fn {download_function}}"
      ))
    }
    return(file_path)
  }

  if (!file.exists(default_name)) {
    cli_abort(c(
      "Data file not found: {.file {default_name}}",
      "i" = "Download data using {.fn {download_function}}",
      "i" = "Or specify custom path with {.arg use_file}"
    ))
  }

  default_name
}

#' Get and Validate API Key
#'
#' @param api_key user-provided API key
#' @param env_vars character vector of environment variable names
#' @param service_name name of service for error messages
#' @keywords internal
#' @noRd
get_api_key <- function(api_key = NULL, env_vars, service_name) {
  if (!is.null(api_key)) {
    assert_character(api_key, len = 1)
    return(api_key)
  }

  for (var in env_vars) {
    key <- Sys.getenv(var)
    if (!identical(key, "")) {
      return(key)
    }
  }

  cli_abort(c(
    "{service_name} API key not found",
    "i" = "Provide via {.arg api_key} parameter",
    "i" = "Or set environment variable: {.envvar {env_vars[1]}}"
  ))
}

#' Build LLM Categorization Prompt
#'
#' @param domain domain name to classify
#' @param categories allowed categories
#' @keywords internal
#' @noRd
build_categorization_prompt <- function(domain, categories) {
  glue(
    "Classify the website domain '{domain}' into exactly one of these categories: ",
    "{paste(categories, collapse = ', ')}. ",
    "Consider what type of content this domain likely hosts. ",
    "Respond with only the category name, no explanation or punctuation."
  )
}

#' Apply Rate Limiting
#'
#' @param i current iteration
#' @param total total iterations
#' @param rate_limit delay in seconds
#' @keywords internal
#' @noRd
apply_rate_limit <- function(i, total, rate_limit) {
  if (i < total && rate_limit > 0) {
    Sys.sleep(rate_limit)
  }
}

test_that("clean_domains removes common prefixes and paths", {
  expect_equal(rdomains:::clean_domains("  google.com  "), "google.com")
  expect_equal(rdomains:::clean_domains("http://www.google.com"), "google.com")
  expect_equal(rdomains:::clean_domains("https://google.com"), "google.com")
  expect_equal(rdomains:::clean_domains("www.google.com"), "google.com")
  expect_equal(rdomains:::clean_domains("https://google.com/path"), "google.com")
  expect_equal(rdomains:::clean_domains("http://www.google.com/path/to/page"), "google.com")
})

test_that("clean_domains handles multiple domains", {
  input <- c("http://www.google.com", "  facebook.com  ", "https://twitter.com/user")
  expected <- c("google.com", "facebook.com", "twitter.com")
  expect_equal(rdomains:::clean_domains(input), expected)
})

test_that("validate_domains catches NULL input", {
  expect_error(
    rdomains:::validate_domains(NULL),
    "must not be NULL"
  )
})

test_that("validate_domains catches empty strings", {
  expect_error(
    rdomains:::validate_domains(""),
    "contains empty strings"
  )
  expect_error(
    rdomains:::validate_domains(c("google.com", "", "facebook.com")),
    "contains empty strings"
  )
})

test_that("validate_domains catches non-character input", {
  expect_error(
    rdomains:::validate_domains(123),
    "character"
  )
  expect_error(
    rdomains:::validate_domains(list("google.com")),
    "character"
  )
})

test_that("validate_domains accepts valid input", {
  expect_silent(rdomains:::validate_domains("google.com"))
  expect_silent(rdomains:::validate_domains(c("google.com", "facebook.com")))
})

test_that("validate_data_file provides helpful errors for missing files", {
  expect_error(
    rdomains:::validate_data_file("nonexistent.csv", "default.csv", "get_data"),
    "File not found"
  )
})

test_that("validate_data_file works with NULL path and existing default", {
  temp_file <- tempfile(fileext = ".csv")
  writeLines("test", temp_file)

  result <- rdomains:::validate_data_file(NULL, temp_file, "get_data")
  expect_equal(result, temp_file)

  unlink(temp_file)
})

test_that("validate_data_file works with specified path", {
  temp_file <- tempfile(fileext = ".csv")
  writeLines("test", temp_file)

  result <- rdomains:::validate_data_file(temp_file, "default.csv", "get_data")
  expect_equal(result, temp_file)

  unlink(temp_file)
})

test_that("get_api_key retrieves from parameter", {
  key <- rdomains:::get_api_key("test_key", "ENV_VAR", "Service")
  expect_equal(key, "test_key")
})

test_that("get_api_key retrieves from environment", {
  Sys.setenv(TEST_API_KEY = "env_key")
  key <- rdomains:::get_api_key(NULL, "TEST_API_KEY", "Service")
  expect_equal(key, "env_key")
  Sys.unsetenv("TEST_API_KEY")
})

test_that("get_api_key errors when key not found", {
  Sys.unsetenv("MISSING_KEY")
  expect_error(
    rdomains:::get_api_key(NULL, "MISSING_KEY", "Service"),
    "API key not found"
  )
})

test_that("build_categorization_prompt creates valid prompt", {
  prompt <- rdomains:::build_categorization_prompt("google.com", c("tech", "social", "other"))
  expect_true(grepl("google.com", prompt))
  expect_true(grepl("tech", prompt))
  expect_true(grepl("social", prompt))
  expect_true(grepl("other", prompt))
})

test_that("apply_rate_limit sleeps appropriately", {
  start_time <- Sys.time()
  rdomains:::apply_rate_limit(1, 3, 0.1)
  elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
  expect_true(elapsed >= 0.1)
})

test_that("apply_rate_limit doesn't sleep on last iteration", {
  start_time <- Sys.time()
  rdomains:::apply_rate_limit(3, 3, 1.0)
  elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
  expect_true(elapsed < 0.5)
})

#' Honour robots.txt
#'
#' The sibling Python package does not do this -- I audited it and found robots.txt is
#' never fetched on its live path, `Crawl-delay` never parsed, and the user-agent
#' impersonates Chrome while a browser flag actively hides the automation. That is a defect
#' rather than a design, so rdomains does not copy it. Its outcome vocabulary already
#' anticipates this: `robots_blocked` exists in `piedomains/outcomes.py` and is simply
#' never emitted for live fetches.
#'
#' Written here rather than taken from the `robotstxt` package for one decisive reason:
#' that package fetches robots.txt with its own HTTP stack, which would bypass the
#' timeouts, byte caps, address guards and throttle this package is built around. The
#' robots fetch is itself a request to a host we have not yet validated.
#'
#' Parsing follows RFC 9309: group by `User-agent`, prefer the most specific matching
#' group over `*`, longest matching rule wins, and ties go to `Allow`.
#'
#' @keywords internal
#' @noRd
NULL

#' RFC 9309 says 500 KiB; anything larger is not a robots file.
#' @keywords internal
#' @noRd
RD_ROBOTS_MAX_BYTES <- 512000L

#' Parse a robots.txt body into rules for one agent
#'
#' @param txt Contents of robots.txt.
#' @param agent Our product token, lowercased.
#'
#' @return A list with `rules` (data frame of `pattern`, `allow`) and `crawl_delay`.
#' @keywords internal
#' @noRd
parse_robots <- function(txt, agent = "rdomains") {
  lines <- strsplit(txt, "\r?\n")[[1]]
  # Strip comments and a UTF-8 BOM.
  lines <- sub("\uFEFF", "", lines, useBytes = TRUE)
  lines <- sub("#.*$", "", lines)
  lines <- str_trim(lines)
  lines <- lines[nzchar(lines)]

  groups <- list()
  current_agents <- character()
  starting_group <- FALSE

  for (line in lines) {
    parts <- strsplit(line, ":", fixed = TRUE)[[1]]
    if (length(parts) < 2) next
    field <- tolower(str_trim(parts[1]))
    value <- str_trim(paste(parts[-1], collapse = ":"))

    if (identical(field, "user-agent")) {
      if (!starting_group) {
        current_agents <- character()
        starting_group <- TRUE
      }
      current_agents <- c(current_agents, tolower(value))
      next
    }
    if (!length(current_agents)) next
    starting_group <- FALSE

    for (a in current_agents) {
      if (is.null(groups[[a]])) {
        groups[[a]] <- list(rules = list(), crawl_delay = NA_real_)
      }
      if (field %in% c("allow", "disallow")) {
        groups[[a]]$rules[[length(groups[[a]]$rules) + 1]] <- list(
          pattern = value, allow = identical(field, "allow")
        )
      } else if (identical(field, "crawl-delay")) {
        delay <- suppressWarnings(as.numeric(value))
        if (!is.na(delay)) groups[[a]]$crawl_delay <- delay
      }
    }
  }

  # Most specific matching group wins; otherwise the wildcard group.
  agent <- tolower(agent)
  pick <- NULL
  for (name in names(groups)) {
    if (name != "*" && grepl(name, agent, fixed = TRUE)) {
      pick <- name
      break
    }
  }
  if (is.null(pick)) pick <- if ("*" %in% names(groups)) "*" else NULL
  if (is.null(pick)) {
    return(list(rules = empty_rules(), crawl_delay = NA_real_))
  }

  chosen <- groups[[pick]]
  if (!length(chosen$rules)) {
    return(list(rules = empty_rules(), crawl_delay = chosen$crawl_delay))
  }
  rules <- data.frame(
    pattern = vapply(chosen$rules, function(r) r$pattern, character(1)),
    allow = vapply(chosen$rules, function(r) r$allow, logical(1)),
    stringsAsFactors = FALSE
  )
  list(rules = rules, crawl_delay = chosen$crawl_delay)
}

#' @keywords internal
#' @noRd
empty_rules <- function() {
  data.frame(pattern = character(), allow = logical(), stringsAsFactors = FALSE)
}

#' Translate a robots path pattern to a regex
#'
#' `*` matches any run of characters and `$` anchors the end. Everything else is literal.
#'
#' @param pattern robots.txt path pattern
#' @return regex string
#' @keywords internal
#' @noRd
robots_pattern_to_regex <- function(pattern) {
  anchored_end <- grepl("\\$$", pattern)
  if (anchored_end) pattern <- sub("\\$$", "", pattern)
  parts <- strsplit(pattern, "*", fixed = TRUE)[[1]]
  if (!length(parts)) parts <- ""
  escaped <- vapply(parts, function(p) gsub("([.\\\\+?^$(){}|\\[\\]])", "\\\\\\1", p),
                    character(1), USE.NAMES = FALSE)
  # A trailing "*" produces no final element from strsplit; re-add the wildcard.
  trailing <- if (grepl("\\*$", pattern)) ".*" else ""
  paste0("^", paste(escaped, collapse = ".*"), trailing, if (anchored_end) "$" else "")
}

#' Is a path allowed by parsed rules?
#'
#' Longest matching pattern wins; a tie goes to Allow, per RFC 9309.
#'
#' @param rules data frame of pattern/allow
#' @param path request path, beginning with "/"
#' @return logical
#' @keywords internal
#' @noRd
robots_path_allowed <- function(rules, path) {
  if (!nrow(rules)) {
    return(TRUE)
  }
  best_len <- -1L
  best_allow <- TRUE
  for (i in seq_len(nrow(rules))) {
    pattern <- rules$pattern[i]
    # An empty Disallow means "allow everything" and matches nothing.
    if (!nzchar(pattern)) next
    if (grepl(robots_pattern_to_regex(pattern), path)) {
      len <- nchar(pattern)
      if (len > best_len || (len == best_len && rules$allow[i])) {
        best_len <- len
        best_allow <- rules$allow[i]
      }
    }
  }
  best_allow
}

#' Fetch, parse and cache robots.txt for one origin
#'
#' Fail-closed on the cases where we cannot know: a 5xx, a timeout or an unparseable body
#' means disallow. A 4xx means allow, which is what RFC 9309 specifies and is by far the
#' common case.
#'
#' @param scheme_host origin, e.g. "https://example.com"
#' @param agent our product token
#' @param timeout seconds
#' @param ttl cache lifetime in seconds
#'
#' @return list(rules, crawl_delay, status, fetched_at)
#' @keywords internal
#' @noRd
robots_for_origin <- function(scheme_host, agent = "rdomains", timeout = 10,
                              ttl = 86400) {
  key <- paste0("robots:", scheme_host)
  hit <- .rdomains_env[[key]]
  if (!is.null(hit) && difftime(Sys.time(), hit$fetched_at, units = "secs") < ttl) {
    return(hit)
  }

  result <- tryCatch({
    resp <- request(paste0(scheme_host, "/robots.txt")) |>
      req_user_agent(agent) |>
      req_timeout(timeout) |>
      req_error(is_error = function(resp) FALSE) |>
      req_perform()
    status <- resp_status(resp)
    if (status >= 500) {
      list(rules = deny_all(), crawl_delay = NA_real_, status = status)
    } else if (status >= 400) {
      list(rules = empty_rules(), crawl_delay = NA_real_, status = status)
    } else {
      body <- resp_body_string(resp)
      if (nchar(body, type = "bytes") > RD_ROBOTS_MAX_BYTES) {
        list(rules = deny_all(), crawl_delay = NA_real_, status = status)
      } else {
        parsed <- parse_robots(body, agent)
        list(rules = parsed$rules, crawl_delay = parsed$crawl_delay, status = status)
      }
    }
  }, error = function(e) {
    list(rules = deny_all(), crawl_delay = NA_real_, status = NA_integer_)
  })

  result$fetched_at <- Sys.time()
  .rdomains_env[[key]] <- result
  result
}

#' @keywords internal
#' @noRd
deny_all <- function() {
  data.frame(pattern = "/", allow = FALSE, stringsAsFactors = FALSE)
}

#' Get Category from University Domain List
#'
#' Fetches university domain json from:
#' \url{https://raw.githubusercontent.com/Hipo/university-domains-list/master/world_universities_and_domains.json}
#'
#' @param domains vector of domain names
#'
#' @return data.frame with original list and all the other columns from the university json
#'
#' @export
#' @importFrom dplyr bind_cols
#' @examples \dontrun{
#' uni_cat(domains = "http://www.google.com")
#' }

uni_cat <- function(domains = NULL) {

  validate_domains(domains)
  c_domains <- clean_domains(domains)

  uni_list <- fromJSON(paste0("https://raw.githubusercontent.com/Hipo/",
                              "university-domains-list/master/",
                              "world_universities_and_domains.json"))

  tibble(
    domain_name = c_domains
  ) |>
    bind_cols(
      uni_list[match(c_domains, uni_list$domains), ]
    )
}

#' Get Category from VirusTotal
#'
#' Returns category of content from multiple security vendors using the VirusTotal API v3.
#' The function retrieves domain analysis results including categories from various security
#' services. Not all services will have categories for all domains.
#'
#' Get the API Access Key from \url{https://www.virustotal.com/}. Either pass the API Key to the function
#' or set the environmental variable: \code{VirustotalToken}. Environment variables persist within
#' a R session.
#'
#' @param domains domain names as character vector
#' @param apikey virustotal API key
#'
#' @return data.frame with domain and VirusTotal analysis results
#'
#' @export
#' @references \url{https://docs.virustotal.com/reference/domains}
#'
#' @examples \dontrun{
#' virustotal_cat("http://www.google.com")
#' virustotal_cat(c("google.com", "facebook.com"))
#' }

virustotal_cat <- function(domains = NULL, apikey = NULL) {

  validate_domains(domains, "domains")

  if (identical(Sys.getenv("VirustotalToken"), "")) {
    if (is.null(apikey)) {
      cli_abort(c(
        "VirusTotal API key not found",
        "i" = "Provide via {.arg apikey} parameter",
        "i" = "Or set environment variable: {.envvar VirustotalToken}"
      ))
    }
    set_key(apikey)
  }

  results <- map_df(domains, function(domain) {
    res <- tryCatch(
      domain_report(domain),
      error = function(e) {
        cli_warn(c(
          "Error processing domain: {domain}",
          "x" = e$message
        ))
        return(data.frame(domain = domain))
      }
    )

    has_categories <- !is.null(res$data) &&
                      !is.null(res$data$attributes$categories) &&
                      length(res$data$attributes$categories) > 0

    if (has_categories) {
      categories <- res$data$attributes$categories
      cat_df <- as.data.frame(as.list(categories), stringsAsFactors = FALSE)
      data.frame(domain = domain, cat_df, stringsAsFactors = FALSE)
    } else {
      data.frame(domain = domain)
    }
  })

  results
}

#' Fetch homepage content for a set of domains
#'
#' Mirrors the sibling Python package's `DataCollector` (`src/piedomains/data_collector.py`)
#' and its outcome contract: **one row per requested domain, never dropped**, each carrying
#' `status`, `stage`, `error_code` and `retryable` so a caller can tell a transient failure
#' from a permanent one and retry exactly the right rows.
#'
#' Static HTML only -- no headless browser. Pages that render their content with JavaScript
#' will come back thin, and `page_state` says so rather than pretending otherwise.
#'
#' **This crawler identifies itself and obeys robots.txt.** That is a deliberate departure
#' from piedomains, which I audited and found does neither on its live path. A CRAN package
#' fetching arbitrary URLs on a user's behalf has no business impersonating a browser.
#'
#' @keywords internal
#' @noRd
NULL

#' Identify ourselves honestly, with a way to be blocked
#' @keywords internal
#' @noRd
rdomains_user_agent <- function() {
  sprintf("rdomains/%s (R package; +https://github.com/themains/rdomains)",
          utils::packageVersion("rdomains"))
}

#' Addresses that must never be fetched
#'
#' Rejecting these stops a domain lookup from being turned into a probe of the caller's own
#' network. `169.254.169.254` in particular is the cloud metadata endpoint.
#'
#' @param ip character IP address
#' @return TRUE when the address is globally routable
#' @keywords internal
#' @noRd
is_global_ip <- function(ip) {
  if (is.na(ip) || !nzchar(ip)) {
    return(FALSE)
  }
  # IPv4-mapped IPv6 (::ffff:10.0.0.1) must be unwrapped before checking -- this is the
  # step people forget, and it re-opens the whole hole.
  mapped <- str_match(tolower(ip), "^::ffff:(\\d+\\.\\d+\\.\\d+\\.\\d+)$")
  if (!is.na(mapped[1, 2])) ip <- mapped[1, 2]

  if (grepl(":", ip, fixed = TRUE)) {
    lowered <- tolower(ip)
    if (lowered %in% c("::", "::1")) return(FALSE)
    # fc00::/7 unique-local, fe80::/10 link-local
    if (grepl("^f[cd]", lowered)) return(FALSE)
    if (grepl("^fe[89ab]", lowered)) return(FALSE)
    return(TRUE)
  }

  octets <- suppressWarnings(as.integer(strsplit(ip, ".", fixed = TRUE)[[1]]))
  if (length(octets) != 4 || any(is.na(octets))) {
    return(FALSE)
  }
  a <- octets[1]; b <- octets[2]
  !(a == 0 || a == 10 || a == 127 ||
      (a == 100 && b >= 64 && b <= 127) ||
      (a == 169 && b == 254) ||
      (a == 172 && b >= 16 && b <= 31) ||
      (a == 192 && b == 168) ||
      (a == 192 && b == 0) ||
      (a == 198 && (b == 18 || b == 19)) ||
      a >= 224)
}

#' Reject a URL we should not fetch
#'
#' @param url the URL
#' @param resolver function(host) -> character vector of IPs; injectable so tests never
#'   touch DNS
#' @return NULL when acceptable, otherwise an error code
#' @keywords internal
#' @noRd
check_url <- function(url, resolver = nslookup) {
  parsed <- tryCatch(url_parse(url), error = function(e) NULL)
  if (is.null(parsed) || is.null(parsed$hostname) || !nzchar(parsed$hostname)) {
    return("invalid_domain")
  }
  if (!identical(parsed$scheme, "http") && !identical(parsed$scheme, "https")) {
    return("invalid_domain")
  }
  host <- tolower(parsed$hostname)
  # A domain classifier has nothing to say about a bare IP, and .onion is unreachable.
  if (grepl("^\\d+\\.\\d+\\.\\d+\\.\\d+$", host) ||
      grepl("\\.(onion|local|internal|localhost)$", host) ||
      !grepl(".", host, fixed = TRUE)) {
    return("invalid_domain")
  }

  ips <- tryCatch(resolver(host, error = FALSE), error = function(e) NULL)
  if (is.null(ips) || !length(ips)) {
    return("dns_error")
  }
  if (!all(vapply(ips, is_global_ip, logical(1)))) {
    return("private_address")
  }
  NULL
}

#' Map a caught condition to a stable error code
#'
#' Matches on condition class first, because R's classes carry better information than
#' message text and matching on the word "timeout" misclassifies a page whose error text
#' merely contains it.
#'
#' @param cnd condition
#' @return error code string
#' @keywords internal
#' @noRd
classify_condition <- function(cnd) {
  classes <- class(cnd)
  msg <- tolower(conditionMessage(cnd))
  if (any(grepl("timeout", classes)) || grepl("timed? ?out", msg)) return("timeout")
  if (any(grepl("httr2_failure", classes))) {
    if (grepl("could not resolve|name or service|nodename", msg)) return("dns_error")
    if (grepl("ssl|certificate|tls", msg)) return("connection_error")
    return("connection_error")
  }
  if (grepl("could not resolve|name or service|nodename", msg)) return("dns_error")
  if (grepl("refused|reset|unreachable", msg)) return("connection_error")
  "unknown"
}

#' Fetch homepage HTML and text for domains
#'
#' Every requested domain comes back, in order, whether or not it was reachable. Failures
#' carry a code from [fetch_error_codes()] rather than a bare `NA`.
#'
#' The crawler identifies itself as `rdomains/<version>`, obeys `robots.txt` including
#' `Crawl-delay`, spaces requests to the same host, caps the response body, and refuses to
#' fetch hosts that resolve to private or link-local addresses.
#'
#' @param domains Character vector of domains or URLs.
#' @param delay Minimum seconds between requests to the same host. `Crawl-delay` overrides
#'   this upward.
#' @param timeout Per-request timeout, seconds.
#' @param max_bytes Cap on the response body actually read.
#' @param obey_robots Whether to fetch and honour robots.txt. Turning this off is
#'   discouraged and is your responsibility, not the package's.
#' @param max_crawl_delay Skip a host that asks for a longer delay than this rather than
#'   sleeping on it.
#' @param max_redirects Maximum redirect hops to follow. Every hop is re-validated, so a
#'   redirect cannot be used to reach an address the first check refused.
#' @param user_agent Override the identifying user-agent.
#'
#' @return A tibble with one row per input: `domain_name`, `status`, `stage`, `error_code`,
#'   `retryable`, `http_status`, `final_url`, `fetched_at`, `content_bytes`, `title`,
#'   `description`, `lang`, `text`, `n_tokens`, `page_state`, `block_vendor`,
#'   `robots_allowed`, `source_last_published`.
#'
#' @export
#' @seealso [fetch_report()] to summarise the run, [page_signals()] for what the page
#'   states are, [source_vintage()] for how a live fetch compares with the static lists.
#' @examples \dontrun{
#' res <- collect_content(c("example.com", "wikipedia.org"))
#' fetch_report(res)
#'
#' # retry only what is worth retrying
#' again <- collect_content(res$domain_name[res$retryable])
#' }
collect_content <- function(domains = NULL,
                            delay = 1,
                            timeout = 10,
                            max_bytes = 2 * 1024^2,
                            obey_robots = TRUE,
                            max_crawl_delay = 30,
                            max_redirects = 5,
                            user_agent = rdomains_user_agent()) {
  validate_domains(domains)
  clean <- clean_domains(domains)

  rows <- lapply(seq_along(clean), function(i) {
    fetch_one(
      clean[i], domains[i],
      delay = delay, timeout = timeout, max_bytes = max_bytes,
      obey_robots = obey_robots, max_crawl_delay = max_crawl_delay,
      max_redirects = max_redirects, user_agent = user_agent
    )
  })
  dplyr::bind_rows(rows)
}

#' Build one result row with a fixed schema and fixed column types
#'
#' R has no static types, so the schema is pinned here instead: every row in a run comes
#' from this one constructor, with every column explicitly coerced. Two separate `tibble()`
#' calls will drift, and when they do `bind_rows()` fails with a type error far from the
#' cause -- or worse, succeeds. This function is the type declaration.
#'
#' @param domain cleaned domain
#' @param input the caller's original string
#' @param started fetch start time
#' @param status "ok" or "failed"
#' @param stage pipeline stage reached
#' @param error_code code from [fetch_error_codes()], or NA
#' @param http_status HTTP status code, or NA
#' @param final_url URL after redirects, or NA
#' @param content_bytes body size actually read
#' @param parsed result of [html_text_content()], or NULL
#' @param signals result of [page_signals()], or NULL
#' @param robots_allowed whether robots.txt permitted the fetch
#'
#' @return a one-row tibble
#' @keywords internal
#' @noRd
fetch_row <- function(domain, input, started,
                      status = "failed", stage = "fetch", error_code = NA_character_,
                      http_status = NA_integer_, final_url = NA_character_,
                      content_bytes = NA_integer_, parsed = NULL, signals = NULL,
                      robots_allowed = NA) {
  tibble(
    domain_name = as.character(domain),
    input = as.character(input),
    status = as.character(status),
    stage = as.character(stage),
    error_code = as.character(error_code),
    retryable = isTRUE(is_retryable(error_code)),
    http_status = as.integer(http_status),
    final_url = as.character(final_url),
    fetched_at = started,
    content_bytes = as.integer(content_bytes),
    title = as.character(parsed$title %||% NA_character_),
    description = as.character(parsed$description %||% NA_character_),
    lang = as.character(parsed$lang %||% NA_character_),
    text = as.character(parsed$text %||% NA_character_),
    n_tokens = as.integer(signals$n_tokens %||% 0L),
    page_state = as.character(signals$page_state %||% NA_character_),
    block_vendor = as.character(signals$block_vendor %||% NA_character_),
    robots_allowed = as.logical(robots_allowed),
    source_last_published = format(started, "%Y-%m")
  )
}

#' @keywords internal
#' @noRd
`%||%` <- function(x, y) if (is.null(x)) y else x

#' Fetch a single domain, returning one fully-populated row whatever happens
#' @keywords internal
#' @noRd
fetch_one <- function(domain, input, delay, timeout, max_bytes, obey_robots,
                      max_crawl_delay, max_redirects, user_agent) {
  started <- Sys.time()
  url <- paste0("https://", domain)

  blank <- function(code, stage, http_status = NA_integer_, robots_allowed = NA) {
    fetch_row(domain, input, started, status = "failed", stage = stage,
              error_code = code, http_status = http_status,
              robots_allowed = robots_allowed)
  }

  bad <- check_url(url)
  if (!is.null(bad)) {
    return(blank(bad, if (bad == "dns_error") "fetch" else "validate"))
  }

  crawl_delay <- delay
  if (obey_robots) {
    origin <- paste0("https://", domain)
    rules <- robots_for_origin(origin, agent = "rdomains", timeout = timeout)
    if (!robots_path_allowed(rules$rules, "/")) {
      return(blank("robots_blocked", "validate", robots_allowed = FALSE))
    }
    if (!is.na(rules$crawl_delay)) {
      if (rules$crawl_delay > max_crawl_delay) {
        return(blank("robots_blocked", "validate", robots_allowed = FALSE))
      }
      crawl_delay <- max(crawl_delay, rules$crawl_delay)
    }
  }

  # Redirects are followed by hand rather than by curl, so every hop goes back through
  # check_url(). Letting curl follow them would validate only the domain the caller asked
  # for: a host redirecting to 127.0.0.1 or 169.254.169.254 would then be fetched, which
  # makes the address guard decorative. Validating the entry URL alone is not an SSRF
  # guard.
  chain <- character()
  current <- url
  resp <- NULL
  for (hop in seq_len(max_redirects + 1L)) {
    resp <- tryCatch({
      request(current) |>
        req_user_agent(user_agent) |>
        req_timeout(timeout) |>
        req_throttle(capacity = 1L, fill_time_s = crawl_delay) |>
        req_retry(max_tries = 2, max_seconds = 30) |>
        # Without this a 403 interstitial throws and page_signals() never sees the body --
        # which is exactly the case we most need to classify.
        req_error(is_error = function(resp) FALSE) |>
        req_options(maxfilesize_large = max_bytes, followlocation = FALSE) |>
        req_perform()
    }, error = function(e) e)

    if (inherits(resp, "condition")) {
      return(blank(classify_condition(resp), "fetch", robots_allowed = TRUE))
    }

    code <- resp_status(resp)
    location <- tryCatch(resp_header(resp, "location"), error = function(e) NULL)
    if (!(code >= 300 && code < 400) || is.null(location) || !nzchar(location)) {
      break
    }
    if (hop > max_redirects) {
      return(blank("too_many_redirects", "fetch", http_status = code,
                   robots_allowed = TRUE))
    }
    current <- url_absolute(location, current)
    chain <- c(chain, current)
    bad_hop <- check_url(current)
    if (!is.null(bad_hop)) {
      return(blank(bad_hop, "validate", http_status = code, robots_allowed = TRUE))
    }
  }

  # Named `http_code`, not `status`. tibble() evaluates its arguments sequentially and
  # later ones see earlier ones, so a `status = ...` column would shadow this variable and
  # `http_status = status` would silently store the string "ok"/"failed" instead of the
  # HTTP code. That exact mistake is why fetch_report() miscounted; it is worth naming.
  http_code <- as.integer(resp_status(resp))
  ctype <- tryCatch(resp_content_type(resp), error = function(e) NA_character_)
  body <- tryCatch(resp_body_string(resp), error = function(e) "")
  bytes <- nchar(body, type = "bytes")

  if (bytes > max_bytes) {
    return(blank("content_too_large", "fetch", http_status = http_code,
                 robots_allowed = TRUE))
  }
  if (!is.na(ctype) && !grepl("html|xml", ctype, fixed = FALSE)) {
    return(blank("content_type_rejected", "fetch", http_status = http_code,
                 robots_allowed = TRUE))
  }

  parsed <- html_text_content(body)
  signals <- page_signals(body, text = parsed$text, domain = domain, status = http_code)

  code <- NA_character_
  state_status <- "ok"
  stage <- "process"
  if (signals$blocked) {
    code <- "bot_blocked"; state_status <- "failed"; stage <- "fetch"
  } else if (http_code >= 400) {
    code <- "http_error"; state_status <- "failed"; stage <- "fetch"
  } else if (!nzchar(parsed$text)) {
    code <- "empty_text"; state_status <- "failed"
  } else if (signals$thin && signals$page_state == "thin") {
    code <- "thin_content"; state_status <- "failed"
  }

  fetch_row(
    domain, input, started,
    status = state_status,
    stage = if (identical(state_status, "ok")) "process" else stage,
    error_code = code,
    http_status = http_code,
    final_url = tryCatch(resp_url(resp), error = function(e) url),
    content_bytes = bytes,
    parsed = parsed,
    signals = signals,
    robots_allowed = TRUE
  )
}

test_that("private and special addresses are never fetchable", {
  # Rejecting these is what stops a domain lookup from becoming a probe of the caller's
  # own network. 169.254.169.254 is the cloud metadata endpoint.
  for (ip in c("127.0.0.1", "10.1.2.3", "192.168.1.1", "169.254.169.254",
               "172.16.0.1", "100.64.0.1", "0.0.0.0", "224.0.0.1",
               "::1", "fe80::1", "fc00::1")) {
    expect_false(is_global_ip(ip), info = ip)
  }
  for (ip in c("8.8.8.8", "1.1.1.1", "93.184.216.34", "2001:4860:4860::8888")) {
    expect_true(is_global_ip(ip), info = ip)
  }
})

test_that("IPv4-mapped IPv6 is unwrapped before the check", {
  # The step people forget, and forgetting it re-opens the whole hole.
  expect_false(is_global_ip("::ffff:10.0.0.1"))
  expect_false(is_global_ip("::ffff:127.0.0.1"))
  expect_true(is_global_ip("::ffff:8.8.8.8"))
})

test_that("is_global_ip rejects garbage rather than defaulting to allow", {
  expect_false(is_global_ip(NA_character_))
  expect_false(is_global_ip(""))
  expect_false(is_global_ip("not-an-ip"))
  expect_false(is_global_ip("999.1.1.1"))
})

test_that("check_url refuses what should never be fetched", {
  # DNS is injected so this never touches the network.
  global <- function(host, error = FALSE) "8.8.8.8"
  private <- function(host, error = FALSE) "10.0.0.1"

  expect_null(check_url("https://example.com", resolver = global))
  expect_equal(check_url("https://example.com", resolver = private), "private_address")
  expect_equal(check_url("ftp://example.com", resolver = global), "invalid_domain")
  expect_equal(check_url("https://1.2.3.4", resolver = global), "invalid_domain")
  expect_equal(check_url("https://secret.onion", resolver = global), "invalid_domain")
  expect_equal(check_url("https://localhost", resolver = global), "invalid_domain")
  expect_equal(check_url("https://box.local", resolver = global), "invalid_domain")
  expect_equal(
    check_url("https://example.com", resolver = function(host, error = FALSE) NULL),
    "dns_error"
  )
})

test_that("conditions map to stable codes by class, not message text", {
  expect_equal(classify_condition(simpleError("Timeout was reached")), "timeout")
  expect_equal(classify_condition(simpleError("Could not resolve host")), "dns_error")
  expect_equal(classify_condition(simpleError("Connection refused")), "connection_error")
  expect_equal(classify_condition(simpleError("something odd")), "unknown")
})

test_that("robots.txt parsing follows RFC 9309", {
  txt <- paste(
    "User-agent: *", "Disallow: /private", "Allow: /private/public",
    "Crawl-delay: 2", "", "User-agent: badbot", "Disallow: /", sep = "\n"
  )
  p <- parse_robots(txt, "rdomains")
  expect_equal(p$crawl_delay, 2)
  expect_true(robots_path_allowed(p$rules, "/"))
  expect_false(robots_path_allowed(p$rules, "/private"))
  # Longest match wins, so the more specific Allow beats the shorter Disallow.
  expect_true(robots_path_allowed(p$rules, "/private/public"))
  expect_false(robots_path_allowed(p$rules, "/private/secret"))

  # A group naming us specifically beats the wildcard group.
  expect_false(robots_path_allowed(parse_robots(txt, "badbot")$rules, "/"))
})

test_that("robots wildcards and end-anchors work", {
  p <- parse_robots("User-agent: *\nDisallow: /*.pdf$\nDisallow: /tmp/")
  expect_false(robots_path_allowed(p$rules, "/a.pdf"))
  expect_true(robots_path_allowed(p$rules, "/a.pdf?x=1"))  # $ anchors the end
  expect_false(robots_path_allowed(p$rules, "/tmp/x"))
  expect_true(robots_path_allowed(p$rules, "/ok.html"))
})

test_that("an empty Disallow allows everything", {
  p <- parse_robots("User-agent: *\nDisallow:")
  expect_true(robots_path_allowed(p$rules, "/anything"))
})

test_that("comments and CRLF do not break parsing", {
  p <- parse_robots("# a comment\r\nUser-agent: *\r\nDisallow: /x  # trailing\r\n")
  expect_false(robots_path_allowed(p$rules, "/x"))
})

test_that("the user agent identifies the package and offers a contact", {
  ua <- rdomains_user_agent()
  expect_match(ua, "^rdomains/")
  expect_match(ua, "github.com/themains/rdomains", fixed = TRUE)
  # Never impersonate a browser: it is dishonest and it does not even work against bot
  # walls, which fingerprint TLS rather than the UA string.
  expect_false(grepl("Mozilla|Chrome|Safari", ua))
})

test_that("collect_content validates its input before touching the network", {
  expect_error(collect_content(NULL), "must not be NULL")
  expect_error(collect_content(""), "contains empty strings")
  expect_error(collect_content(123), "character")
})

test_that("collect_content returns a row per input, even for unfetchable ones", {
  skip_on_cran()
  skip_if_offline()

  res <- collect_content(
    c("example.com", "no-such-domain-xyzzy-99999.invalid"),
    obey_robots = FALSE
  )
  expect_equal(nrow(res), 2)
  expect_equal(res$domain_name, c("example.com", "no-such-domain-xyzzy-99999.invalid"))
  expect_equal(res$error_code[2], "dns_error")
  expect_true(res$retryable[2])
  # Every row carries its own provenance, whatever happened.
  expect_false(any(is.na(res$fetched_at)))
  expect_false(any(is.na(res$source_last_published)))
})

test_that("http_status holds the HTTP code, not the row status", {
  # Regression: tibble() evaluates arguments sequentially, so a `status` column shadowed
  # the local variable holding the HTTP code and http_status silently became "ok"/"failed".
  skip_on_cran()
  skip_if_offline()

  res <- collect_content("example.com", obey_robots = FALSE)
  expect_type(res$http_status, "integer")
  expect_equal(res$http_status[1], 200L)
})

test_that("every row has the same schema and types, whatever happened", {
  # The typing guarantee. R has no static types, so one constructor pins the schema --
  # two separate tibble() calls drift, and when they do bind_rows() either fails far from
  # the cause or silently coerces. This is what caught http_status becoming a character.
  started <- Sys.time()
  ok <- fetch_row(
    "a.com", "a.com", started, status = "ok", stage = "process",
    http_status = 200L, final_url = "https://a.com", content_bytes = 100L,
    parsed = list(title = "T", description = NA_character_, lang = "en", text = "words"),
    signals = list(n_tokens = 5L, page_state = "content", block_vendor = NA_character_),
    robots_allowed = TRUE
  )
  failed <- fetch_row("b.com", "b.com", started, error_code = "dns_error")

  expect_identical(names(ok), names(failed))
  expect_identical(
    vapply(ok, function(x) class(x)[1], character(1)),
    vapply(failed, function(x) class(x)[1], character(1))
  )
  expect_type(ok$http_status, "integer")
  expect_type(ok$retryable, "logical")
  # bind_rows must not error, which is the whole point.
  expect_no_error(dplyr::bind_rows(ok, failed))
})

test_that("retryable is derived from the code, never set by hand", {
  started <- Sys.time()
  expect_true(fetch_row("a.com", "a.com", started, error_code = "timeout")$retryable)
  expect_false(fetch_row("a.com", "a.com", started, error_code = "bot_blocked")$retryable)
  expect_false(fetch_row("a.com", "a.com", started, status = "ok")$retryable)
})

test_that("redirects are followed and every hop is re-validated", {
  # Validating only the entry URL is not an SSRF guard: a host redirecting to
  # 127.0.0.1 or 169.254.169.254 would sail past it. curl follows redirects
  # internally, so following them by hand is what makes the check apply per hop.
  skip_on_cran()
  skip_if_offline()

  res <- collect_content("bit.ly", obey_robots = FALSE)
  expect_equal(res$status, "ok")
  # bit.ly redirects cross-domain; the final URL must reflect that, not the request.
  expect_match(res$final_url, "bitly", fixed = TRUE)
})

test_that("too_many_redirects is in the taxonomy and not retryable", {
  expect_true("too_many_redirects" %in% fetch_error_codes()$code)
  expect_false(is_retryable("too_many_redirects"))
})

#' ML Model
#'
#'
#' @docType data
#' @keywords keywords model
#' @name glm_shalla
#' @source ML model based on shallalist using keywords and domain suffixes,
#'
#' @format A list
#' @author Gaurav Sood
#' @export
"glm_shalla"

#' Probability that Domain Hosts Adult Content Based on features of Domain Name and Suffix alone.
#'
#' Uses a validated ML model that uses keywords in the domain name
#' and suffix to predict probability that the domain hosts adult content. For
#' more information see \url{https://github.com/themains/keyword_porn}
#'
#' @param domains required; string; vector of domain names
#'
#' @return data.frame with original list and content category of the domains
#'
#' @export
#' @examples \dontrun{
#' adult_ml1_cat("http://www.google.com")
#' }

adult_ml1_cat <- function(domains = NULL) {

  validate_domains(domains)
  c_domains <- clean_domains(domains)

  coefs <- dimnames(glm_shalla$glmnet.fit$beta)[[1]]

  features <- spMatrix(length(c_domains), length(coefs))

  for (j in 1:60) {
    tfs <- grepl(coefs[j], c_domains)
    features[, j] <- as(tfs, "sparseVector")
  }

  tfs <- grepl("^[0-9]*.[0-9]*.[0-9]*.[0-9]", c_domains)
  features[, 61] <- as(tfs, "sparseVector")

  split_url <- suffix_extract(c_domains)
  suffixes <- split_url$suffix[match(c_domains, split_url$host)]

  for (t in 62:length(coefs)) {
    tfs <- grepl(coefs[t], suffixes)
    features[, t] <- as(tfs, "sparseVector")
  }

  p_adult <- predict(glm_shalla, features,
                     s = "lambda.min",
                     type = "response")[, 1]

  tibble(
    domain_name = c_domains,
    p_adult = p_adult
  )
}

test_that("Adult ML1 cat", {

  report <- adult_ml1_cat("http://www.google.com")
  expect_s3_class(report, "data.frame")
})

test_that("adult_ml1_cat validates input", {
  expect_error(adult_ml1_cat(NULL), "must not be NULL")
  expect_error(adult_ml1_cat(""), "contains empty strings")
  expect_error(adult_ml1_cat(123), "character")
})

#' Get Category from Shallalist
#'
#' Fetches category of content hosted by a domain according to Shalla.
#' The function checks if path to the shalla file is provided by the user.
#' If not, it looks for \code{shalla_domain_category.csv} in the working directory.
#'
#' Shallalist stopped publishing in 2022, so these labels describe the domain as it was
#' then or earlier, not as it is now. The returned \code{source_last_published} column
#' carries that vintage, and a warning is issued once per session. See
#' \code{\link{source_vintage}}.
#'
#' @param domains vector of domain names
#' @param use_file path to the latest shallalist file downloaded using \code{\link{get_shalla_data}}
#'
#' @return data.frame with the original list, the content category of the domain, and the
#'   vintage of the list that supplied it
#'
#' @export
#' @seealso \code{\link{source_vintage}} for the provenance of every category source
#' @examples \dontrun{
#' shalla_cat(domains = "http://www.google.com")
#' }

shalla_cat <- function(domains = NULL, use_file = NULL) {

  validate_domains(domains)
  c_domains <- clean_domains(domains)

  data_file <- validate_data_file(
    use_file,
    "shalla_domain_category.csv",
    "get_shalla_data"
  )

  shalla <- read_csv(data_file, show_col_types = FALSE)

  warn_source_vintage("shalla")

  tibble(
    domain_name = c_domains,
    shalla_category = shalla$category[match(c_domains, shalla$domains)],
    source_last_published = SOURCE_VINTAGE$shalla$last_published
  )
}

#' Classify News and Non-News Based on keywords in the URL
#'
#' Based on a slightly amended version of the regular expression used to classify news, and non-news in:
#' ``Exposure to ideologically diverse news and opinion on Facebook''
#' by Bakshy, Messing, and Adamic. Science. 2015.
#'
#' Amendment: sport rather than sports
#'
#' URL containing any of the following words is classified as soft news:
#' "sport|entertainment|arts|fashion|style|lifestyle|leisure|celeb|movie|music|gossip|food|travel|horoscope|weather|gadget"
#'
#' URL containing any of following words is classified as hard news:
#' "politi|usnews|world|national|state|elect|vote|govern|campaign|war|polic|econ|unemploy|racis|energy|abortion|educa|healthcare|immigration"
#'
#'
#' Note that it is based on patterns existing in a small set of domains. See paper for details.
#'
#' @param url_list vector of URLs
#'
#' @return data.frame with 3 columns: url, not_news, news
#'
#' @export
#' @references \url{https://www.science.org/doi/10.1126/science.aaa1160}
#'
#' @examples \dontrun{
#' not_news("http://www.bbc.com/sport")
#' not_news(c("http://www.bbc.com/sport", "http://www.washingtontimes.com/news/politics/"))
#' }

not_news <- function(url_list = NULL) {

  validate_domains(url_list, "url_list")

  soft_news_pattern <- paste0(
    "sport|entertainment|arts|fashion|style|lifestyle|leisure|celeb|",
    "movie|music|gossip|food|travel|horoscope|weather|gadget"
  )

  hard_news_pattern <- paste0(
    "politi|usnews|world|national|state|elect|vote|govern|campaign|war|",
    "polic|econ|unemploy|racis|energy|abortion|educa|healthcare|immigration"
  )

  tibble(
    url_list = url_list,
    not_news = str_detect(url_list, stringr::regex(soft_news_pattern, ignore_case = TRUE)),
    news = str_detect(url_list, stringr::regex(hard_news_pattern, ignore_case = TRUE))
  )
}

#' Provenance for the category sources
#'
#' A label has a date. Serving it without one is the bug this file exists to fix.
#'
#' Two of the lookup sources in this package are no longer published: DMOZ closed in
#' March 2017 and Shallalist stopped in 2022. Their labels were correct when they were
#' assigned, but domains expire and get re-registered, so a lookup today can return a
#' category that describes a site which no longer exists. Before this, that answer was
#' presented identically to one from a list updated last week.
#'
#' The sibling project piedomains measured the cost of exactly this confusion: its worst
#' class disagreed with its own page content 71% of the time, and the cause was not bad
#' annotation but a decade between the label and the page. Only 60% of the domains it
#' trained on still resolve.
#'
#' @keywords internal
#' @noRd
.rdomains_env <- new.env(parent = emptyenv())

#' Known vintage of each category source
#'
#' @keywords internal
#' @noRd
SOURCE_VINTAGE <- list(
  shalla = list(
    # get_shalla_data() preserves the final release, dated 2022-01-14.
    source = "Shallalist",
    last_published = "2022-01",
    status = "discontinued",
    note = paste(
      "Shallalist stopped publishing in 2022 and individual entries may be",
      "considerably older. A domain that has changed hands since will carry the",
      "previous registrant's category."
    ),
    successor = "UT-Capitole blacklists (https://dsi.ut-capitole.fr/blacklists/)"
  ),
  dmoz = list(
    source = "DMOZ / Open Directory Project",
    last_published = "2017-03",
    status = "discontinued",
    note = paste(
      "DMOZ closed in March 2017. These labels are at least that old, and the",
      "archived snapshot this package ships is from 2015."
    ),
    successor = "Curlie (https://curlie.org)"
  ),
  stevenblack = list(
    source = "Steven Black unified hosts",
    last_published = NA_character_,
    status = "maintained",
    note = "Updated frequently; the returned date is the fetched file's own.",
    successor = NA_character_
  ),
  uni = list(
    source = "Hipo university-domains-list",
    last_published = NA_character_,
    status = "maintained",
    note = "Community-maintained; no formal release cadence.",
    successor = NA_character_
  )
)

#' Report what each category source is and when it was last published
#'
#' Every lookup in this package answers from a static list, and those lists have very
#' different vintages. Two of them are no longer published at all. Use this to see which
#' answers can be trusted to describe a domain as it is today.
#'
#' @param source Optional; one of \code{"shalla"}, \code{"dmoz"}, \code{"stevenblack"},
#'   \code{"uni"}. If \code{NULL}, every source is returned.
#'
#' @return A tibble with one row per source: \code{key}, \code{source},
#'   \code{last_published}, \code{status}, \code{successor} and \code{note}.
#'
#' @export
#' @examples
#' source_vintage()
#' source_vintage("dmoz")
source_vintage <- function(source = NULL) {
  keys <- names(SOURCE_VINTAGE)
  if (!is.null(source)) {
    assert_character(source, len = 1)
    if (!source %in% keys) {
      cli_abort(c(
        "Unknown source: {.val {source}}",
        "i" = "Known sources: {.val {keys}}"
      ))
    }
    keys <- source
  }

  map_df(keys, function(key) {
    entry <- SOURCE_VINTAGE[[key]]
    tibble(
      key = key,
      source = entry$source,
      last_published = entry$last_published,
      status = entry$status,
      successor = entry$successor,
      note = entry$note
    )
  })
}

#' Warn once per session that a source is no longer published
#'
#' Once per session rather than once per call: a warning on every lookup is a warning
#' people learn to filter out, which leaves them no better informed than no warning at all.
#'
#' @param source key into \code{SOURCE_VINTAGE}
#' @return invisibly, whether a warning was emitted
#' @keywords internal
#' @noRd
warn_source_vintage <- function(source) {
  entry <- SOURCE_VINTAGE[[source]]
  if (is.null(entry) || !identical(entry$status, "discontinued")) {
    return(invisible(FALSE))
  }

  flag <- paste0("warned_", source)
  if (isTRUE(.rdomains_env[[flag]])) {
    return(invisible(FALSE))
  }
  .rdomains_env[[flag]] <- TRUE

  cli_warn(c(
    "{entry$source} is no longer published (last: {entry$last_published}).",
    "!" = entry$note,
    "i" = if (!is.na(entry$successor)) "Successor: {entry$successor}" else NULL,
    "i" = "See {.fn source_vintage}. This warning appears once per session."
  ))
  invisible(TRUE)
}

#' Reset the once-per-session warning state
#'
#' Exists so tests can assert the warning fires, and fires only once.
#'
#' @keywords internal
#' @noRd
reset_vintage_warnings <- function() {
  rm(list = ls(envir = .rdomains_env), envir = .rdomains_env)
  invisible(NULL)
}

test_that("source_vintage lists every source", {
  v <- source_vintage()
  expect_s3_class(v, "data.frame")
  expect_setequal(v$key, c("shalla", "dmoz", "stevenblack", "uni"))
  expect_true(all(c("source", "last_published", "status", "note") %in% names(v)))
})

test_that("source_vintage names the two dead lists", {
  v <- source_vintage()
  dead <- v$key[v$status == "discontinued"]
  expect_setequal(dead, c("shalla", "dmoz"))
  expect_equal(source_vintage("dmoz")$last_published, "2017-03")
  expect_equal(source_vintage("shalla")$last_published, "2022-01")
})

test_that("source_vintage rejects an unknown source", {
  expect_error(source_vintage("curlie"), "Unknown source")
})

test_that("the staleness warning fires once per session, not once per call", {
  reset_vintage_warnings()
  expect_warning(warn_source_vintage("dmoz"), "no longer published")
  # Second call is silent: a warning on every lookup is one people learn to ignore.
  expect_silent(warn_source_vintage("dmoz"))
  expect_silent(warn_source_vintage("dmoz"))
})

test_that("maintained sources never warn", {
  reset_vintage_warnings()
  expect_silent(warn_source_vintage("stevenblack"))
  expect_silent(warn_source_vintage("uni"))
})

test_that("shalla_cat reports the vintage of the label it returns", {
  reset_vintage_warnings()
  tmp <- withr::local_tempfile(fileext = ".csv")
  utils::write.csv(
    data.frame(domains = c("example.com", "test.org"), category = c("shopping", "news")),
    tmp,
    row.names = FALSE
  )

  res <- suppressWarnings(shalla_cat("http://www.example.com", use_file = tmp))
  expect_equal(res$shalla_category, "shopping")
  expect_equal(res$source_last_published, "2022-01")
})

test_that("shalla_cat warns that Shallalist is no longer published", {
  reset_vintage_warnings()
  tmp <- withr::local_tempfile(fileext = ".csv")
  utils::write.csv(
    data.frame(domains = "example.com", category = "shopping"), tmp, row.names = FALSE
  )
  expect_warning(shalla_cat("example.com", use_file = tmp), "no longer published")
})

test_that("dmoz_cat reports its 2017 vintage", {
  reset_vintage_warnings()
  tmp <- withr::local_tempfile(fileext = ".csv")
  writeLines(c("example.com,Shopping", "test.org,News"), tmp)

  res <- suppressWarnings(dmoz_cat("http://www.example.com", use_file = tmp))
  expect_equal(res$dmoz_category, "Shopping")
  expect_equal(res$source_last_published, "2017-03")
})

test_that("an unmatched domain still carries the source vintage", {
  reset_vintage_warnings()
  tmp <- withr::local_tempfile(fileext = ".csv")
  utils::write.csv(
    data.frame(domains = "example.com", category = "shopping"), tmp, row.names = FALSE
  )
  res <- suppressWarnings(shalla_cat("nowhere-at-all.invalid", use_file = tmp))
  expect_true(is.na(res$shalla_category))
  # The label is missing; the provenance of the list that failed to supply it is not.
  expect_equal(res$source_last_published, "2022-01")
})

test_that("dmoz cat", {
  # Downloads an 80 MB archive: CRAN forbids that, and it fails offline.
  skip_on_cran()
  skip_if_offline()

  # Use temporary directory to avoid leaving large files
  temp_dir <- tempdir()
  get_dmoz_data(outdir = temp_dir, overwrite = TRUE)

  # Test with the downloaded data
  data_file <- file.path(temp_dir, "dmoz_domain_category.csv")
  reset_vintage_warnings()
  expect_warning(
    report <- dmoz_cat("http://www.google.com", use_file = data_file),
    "no longer published"
  )

  expect_s3_class(report, "data.frame")
  expect_equal(report$source_last_published, "2017-03")

  # Clean up
  if (file.exists(data_file)) {
    unlink(data_file)
  }
})

test_that("get_dmoz_data writes to outdir, not a hidden file beside it", {
  # paste0(outdir, "name") produced ".dmoz_domain_category.csv" for the documented
  # default outdir = ".", which dmoz_cat() would then fail to find.
  temp_dir <- tempdir()
  expected <- file.path(temp_dir, "dmoz_domain_category.csv")
  writeLines("example.com,Shopping", expected)
  on.exit(unlink(expected), add = TRUE)

  # With the path built correctly, the existing-file guard fires.
  expect_error(
    get_dmoz_data(outdir = temp_dir, overwrite = FALSE),
    "already exists"
  )
})

#' Classify Domain Using Claude (Internal Helper)
#'
#' @param domain single domain name
#' @param categories allowed categories
#' @param model Claude model name
#' @param api_key API key
#' @keywords internal
#' @noRd
classify_domain_claude <- function(domain, categories, model, api_key) {
  prompt <- build_categorization_prompt(domain, categories)

  request_body <- list(
    model = model,
    max_tokens = 50,
    messages = list(
      list(role = "user", content = prompt)
    ),
    temperature = 0.1
  )

  result <- tryCatch({
    response <- httr::POST(
      url = "https://api.anthropic.com/v1/messages",
      httr::add_headers(
        "x-api-key" = api_key,
        "Content-Type" = "application/json",
        "anthropic-version" = "2023-06-01"
      ),
      body = jsonlite::toJSON(request_body, auto_unbox = TRUE),
      encode = "raw"
    )

    if (status_code(response) == 200) {
      result <- fromJSON(content(response, "text", encoding = "UTF-8"))
      category <- str_trim(result$content[1, ]$text)

      if (category %in% categories) {
        category
      } else {
        "other"
      }
    } else {
      cli_warn("API call failed for domain {domain} - Status: {status_code(response)}")
      NA_character_
    }
  }, error = function(e) {
    cli_warn(c(
      "Error processing domain: {domain}",
      "x" = e$message
    ))
    NA_character_
  })

  tibble(domain_name = domain, claude_category = result)
}

#' Get Category from Anthropic Claude
#'
#' Fetches category of content hosted by a domain using Anthropic's Claude API.
#' The function uses Claude models to classify domains into specified categories.
#'
#' @param domains vector of domain names
#' @param api_key Anthropic API key. If not provided, looks for ANTHROPIC_API_KEY or CLAUDE_API_KEY environment variable
#' @param categories vector of categories to classify into. If NULL, uses default web categories
#' @param model Claude model to use (default: "claude-3-haiku-20240307" for cost efficiency)
#' @param rate_limit delay in seconds between API calls (default: 0.5)
#'
#' @return data.frame with original list and content category of the domain
#'
#' @export
#' @examples \dontrun{
#' claude_cat("google.com")
#' claude_cat(c("google.com", "facebook.com"))
#' claude_cat("google.com", categories = c("search", "social", "ecommerce", "news", "other"))
#' }

claude_cat <- function(domains = NULL, api_key = NULL, categories = NULL,
                      model = "claude-3-haiku-20240307", rate_limit = 0.5) {

  validate_domains(domains)
  c_domains <- clean_domains(domains)

  if (is.null(categories)) {
    categories <- c("news", "shopping", "social", "adult", "gambling", "technology",
                   "finance", "education", "entertainment", "business", "other")
  }

  api_key <- get_api_key(api_key, c("ANTHROPIC_API_KEY", "CLAUDE_API_KEY"), "Anthropic")

  map_df(
    seq_along(c_domains),
    function(i) {
      result <- classify_domain_claude(c_domains[i], categories, model, api_key)
      apply_rate_limit(i, length(c_domains), rate_limit)
      result
    }
  )
}

context("Get Claude Cat")

test_that("claude cat validates input", {
  expect_error(claude_cat(), "must not be NULL")

  # Temporarily unset API keys to test error handling
  old_key1 <- Sys.getenv("ANTHROPIC_API_KEY")
  old_key2 <- Sys.getenv("CLAUDE_API_KEY")
  Sys.unsetenv("ANTHROPIC_API_KEY")
  Sys.unsetenv("CLAUDE_API_KEY")
  expect_error(claude_cat("google.com"), "API key not found")
  # Restore the keys
  if (old_key1 != "") Sys.setenv(ANTHROPIC_API_KEY = old_key1)
  if (old_key2 != "") Sys.setenv(CLAUDE_API_KEY = old_key2)
})

test_that("claude cat returns correct structure", {
  skip_if(identical(Sys.getenv("ANTHROPIC_API_KEY"), "") && identical(Sys.getenv("CLAUDE_API_KEY"), ""), 
          "ANTHROPIC_API_KEY or CLAUDE_API_KEY not set")
  
  result <- claude_cat("google.com")
  expect_is(result, "data.frame")
  expect_named(result, c("domain_name", "claude_category"))
  expect_equal(nrow(result), 1)
  expect_equal(result$domain_name[1], "google.com")
})

test_that("claude cat handles multiple domains", {
  skip_if(identical(Sys.getenv("ANTHROPIC_API_KEY"), "") && identical(Sys.getenv("CLAUDE_API_KEY"), ""), 
          "ANTHROPIC_API_KEY or CLAUDE_API_KEY not set")
  
  domains <- c("google.com", "facebook.com")
  result <- claude_cat(domains)
  expect_is(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_equal(result$domain_name, domains)
})

test_that("claude cat preprocesses domains correctly", {
  skip_if(identical(Sys.getenv("ANTHROPIC_API_KEY"), "") && identical(Sys.getenv("CLAUDE_API_KEY"), ""), 
          "ANTHROPIC_API_KEY or CLAUDE_API_KEY not set")
  
  result <- claude_cat("https://www.facebook.com")
  expect_equal(result$domain_name[1], "facebook.com")
})

test_that("claude cat accepts custom categories", {
  skip_if(identical(Sys.getenv("ANTHROPIC_API_KEY"), "") && identical(Sys.getenv("CLAUDE_API_KEY"), ""), 
          "ANTHROPIC_API_KEY or CLAUDE_API_KEY not set")
  
  custom_cats <- c("search", "social", "other")
  result <- claude_cat("facebook.com", categories = custom_cats)
  expect_is(result, "data.frame")
  # Category should be one of the custom categories or NA
  expect_true(is.na(result$claude_category[1]) || result$claude_category[1] %in% custom_cats)
})

#' Why a domain produced no category
#'
#' Ported from the sibling Python package piedomains (`src/piedomains/outcomes.py`), and
#' the strings are **copied verbatim** so results from either package aggregate together.
#' They are part of the public output contract: safe to group on, and they will not change
#' meaning between releases.
#'
#' The problem this solves in rdomains: every failure used to collapse to `NA` plus a
#' warning, so "the host does not resolve", "we were rate-limited" and "this domain has no
#' category" were indistinguishable. Only the second is worth retrying, and a caller had no
#' way to tell which rows those were.
#'
#' @keywords internal
#' @noRd
NULL

#' Pipeline stage a domain reached
#' @keywords internal
#' @noRd
RD_STAGES <- c("validate", "fetch", "process", "infer")

#' The closed set of failure reasons
#'
#' `retryable` marks transient conditions. Two deliberate divergences from piedomains,
#' both because rdomains has no archive.org fallback to retry *into*:
#'   * `bot_blocked` is not retryable here -- retrying from the same address is rudeness
#'     with a guaranteed identical outcome.
#'   * The archive-specific codes are absent.
#'
#' @keywords internal
#' @noRd
RD_ERROR_CODES <- tribble(
  ~code,                   ~stage,     ~retryable, ~description,
  "invalid_domain",        "validate", FALSE, "Not a fetchable http(s) host.",
  "robots_blocked",        "validate", FALSE, "Disallowed by the host's robots.txt.",
  "private_address",       "validate", FALSE, "Host resolves to a non-global address.",
  "dns_error",             "fetch",    TRUE,  "Host did not resolve.",
  "connection_error",      "fetch",    TRUE,  "Connection refused or reset.",
  "timeout",               "fetch",    TRUE,  "Request exceeded the timeout.",
  "http_error",            "fetch",    TRUE,  "Non-2xx status.",
  "content_type_rejected", "fetch",    FALSE, "Response was not HTML.",
  "content_too_large",     "fetch",    FALSE, "Body exceeded the byte cap.",
  "bot_blocked",           "fetch",    FALSE, "An anti-bot interstitial was served.",
  "too_many_redirects",    "fetch",    FALSE, "Redirect chain exceeded max_redirects.",
  "empty_text",            "process",  FALSE, "No text could be extracted.",
  "thin_content",          "process",  FALSE, "Below the token floor for an honest label.",
  "service_unavailable",   "infer",    TRUE,  "Classification service asleep or down.",
  "rate_limited",          "infer",    TRUE,  "Classification service rate-limited us.",
  "service_error",         "infer",    TRUE,  "Classification service returned an error.",
  "unknown",               NA,         FALSE, "Unclassified failure."
)

#' Reasons a domain can fail to get a category
#'
#' The closed set of `error_code` values that [collect_content()] and [pie_cat()] can
#' return, with the pipeline stage each belongs to and whether it is worth retrying.
#'
#' These strings are shared verbatim with the Python package `piedomains`, so results from
#' the two can be aggregated together.
#'
#' @return A tibble with columns `code`, `stage`, `retryable` and `description`.
#'
#' @export
#' @seealso [fetch_report()] to summarise an actual run, [source_vintage()] for the
#'   provenance of the static label sources.
#' @examples
#' fetch_error_codes()
#'
#' # which failures are worth trying again?
#' subset(fetch_error_codes(), retryable)
fetch_error_codes <- function() {
  RD_ERROR_CODES
}

#' Summarise the outcome of a fetch or classification run
#'
#' A pure function over the returned tibble, deliberately **not** a stored summary. A
#' snapshot goes stale the moment you filter the rows; recomputing means
#' `fetch_report(subset(x, retryable))` is always correct.
#'
#' @param x A tibble from [collect_content()] or [pie_cat()].
#'
#' @return A one-row tibble: `n`, `ok`, `failed`, `retryable`, and list-columns
#'   `by_stage`, `by_reason` and `missing` (the domains that produced nothing).
#'
#' @export
#' @examples
#' rows <- data.frame(
#'   domain_name = c("a.com", "b.com", "c.com"),
#'   status = c("ok", "failed", "failed"),
#'   stage = c("infer", "fetch", "process"),
#'   error_code = c(NA, "timeout", "thin_content"),
#'   retryable = c(FALSE, TRUE, FALSE)
#' )
#' fetch_report(rows)
fetch_report <- function(x) {
  if (!is.data.frame(x)) {
    cli_abort("{.arg x} must be a data frame returned by {.fn collect_content}")
  }
  needed <- c("status", "stage", "error_code")
  missing_cols <- setdiff(needed, names(x))
  if (length(missing_cols)) {
    cli_abort("{.arg x} is missing column{?s}: {.field {missing_cols}}")
  }

  # Named `is_failed`, not `failed`: tibble() evaluates its arguments sequentially and
  # later ones see earlier ones, so a `failed = sum(failed)` column would shadow the mask
  # and every subsequent subset would index with the count instead.
  is_failed <- x$status %in% "failed"
  tibble(
    n = nrow(x),
    ok = sum(x$status %in% "ok"),
    failed = sum(is_failed),
    retryable = if ("retryable" %in% names(x)) sum(x$retryable %in% TRUE) else NA_integer_,
    by_stage = list(count_values(x$stage[is_failed])),
    by_reason = list(count_values(x$error_code[is_failed])),
    missing = list(
      if ("domain_name" %in% names(x)) x$domain_name[is_failed] else character()
    )
  )
}

#' Tally a character vector, commonest first
#'
#' @param values character vector, `NA`s dropped
#' @return tibble of value and n
#' @keywords internal
#' @noRd
count_values <- function(values) {
  values <- values[!is.na(values)]
  if (!length(values)) {
    return(tibble(value = character(), n = integer()))
  }
  counted <- sort(table(values), decreasing = TRUE)
  tibble(value = names(counted), n = as.integer(counted))
}

#' Look up whether a code is retryable
#'
#' @param code error code, or NA
#' @return logical
#' @keywords internal
#' @noRd
is_retryable <- function(code) {
  if (is.na(code)) {
    return(FALSE)
  }
  hit <- RD_ERROR_CODES$retryable[match(code, RD_ERROR_CODES$code)]
  isTRUE(hit)
}

test_that("the error-code taxonomy is closed and well formed", {
  codes <- fetch_error_codes()
  expect_s3_class(codes, "data.frame")
  expect_true(all(c("code", "stage", "retryable", "description") %in% names(codes)))
  expect_false(any(duplicated(codes$code)))
  expect_type(codes$retryable, "logical")
  expect_false(any(is.na(codes$retryable)))
  # Every stage named is a real pipeline stage (NA allowed for `unknown`).
  expect_true(all(stats::na.omit(codes$stage) %in% RD_STAGES))
})

test_that("the codes match piedomains verbatim", {
  # These strings are a shared public contract across the two packages. Changing one
  # silently breaks aggregation across them.
  codes <- fetch_error_codes()$code
  shared <- c("invalid_domain", "dns_error", "connection_error", "timeout",
              "http_error", "robots_blocked", "content_type_rejected",
              "content_too_large", "empty_text", "bot_blocked", "thin_content",
              "unknown")
  expect_true(all(shared %in% codes))
})

test_that("bot_blocked is not retryable here, unlike piedomains", {
  # piedomains retries it because it has an archive.org fallback to retry *into*.
  # rdomains has none, so a retry from the same address is rudeness with a guaranteed
  # identical outcome.
  expect_false(is_retryable("bot_blocked"))
  expect_true(is_retryable("timeout"))
  expect_true(is_retryable("dns_error"))
  expect_false(is_retryable("thin_content"))
  expect_false(is_retryable(NA))
})

test_that("fetch_report counts by stage and reason", {
  rows <- tibble::tibble(
    domain_name = c("a.com", "b.com", "c.com", "d.com"),
    status = c("ok", "failed", "failed", "failed"),
    stage = c("infer", "fetch", "fetch", "process"),
    error_code = c(NA, "timeout", "timeout", "thin_content"),
    retryable = c(FALSE, TRUE, TRUE, FALSE)
  )
  rep <- fetch_report(rows)
  expect_equal(rep$n, 4)
  expect_equal(rep$ok, 1)
  expect_equal(rep$failed, 3)
  expect_equal(rep$retryable, 2)
  expect_equal(rep$by_reason[[1]]$value[1], "timeout")
  expect_equal(rep$by_reason[[1]]$n[1], 2)
  expect_setequal(rep$missing[[1]], c("b.com", "c.com", "d.com"))
})

test_that("fetch_report is a pure function, so it survives filtering", {
  # This is why the report is recomputed rather than stored: a stored summary goes
  # stale the moment the rows are subset.
  rows <- tibble::tibble(
    domain_name = c("a.com", "b.com"),
    status = c("ok", "failed"),
    stage = c("infer", "fetch"),
    error_code = c(NA, "timeout"),
    retryable = c(FALSE, TRUE)
  )
  expect_equal(fetch_report(rows[rows$status == "failed", ])$n, 1)
  expect_equal(fetch_report(rows[rows$status == "failed", ])$failed, 1)
})

test_that("fetch_report rejects something that is not a run", {
  expect_error(fetch_report("nope"), "must be a data frame")
  expect_error(fetch_report(tibble::tibble(a = 1)), "missing column")
})

#' Get DMOZ Data
#'
#' Downloads archived DMOZ (Open Directory Project) data. DMOZ was discontinued in March 2017.
#' This function downloads our preserved copy of the final DMOZ dataset. For more details, check:
#' \url{https://github.com/themains/rdomains/tree/master/data-raw/dmoz/}
#'
#' @param outdir    Optional; folder to which you want to save the file; Default is same folder
#' @param overwrite Optional; default is FALSE. If TRUE, the file is overwritten.
#'
#' @export
#'
#' @references \url{https://archive.org/details/dmoz-rdf-20150327}
#'
#' @examples \dontrun{
#' get_dmoz_data()
#' }

get_dmoz_data <- function(outdir = ".", overwrite = FALSE) {

  # file.path, not paste0: with the documented default of outdir = "." the old form
  # produced ".dmoz_domain_category.csv" -- a hidden file that the matching dmoz_cat()
  # lookup would then fail to find. get_shalla_data() and get_stevenblack_data() were
  # already correct.
  dmoz_file <- file.path(outdir, "dmoz_domain_category.csv")

  if (file.exists(dmoz_file) & overwrite == FALSE) {
    stop(paste0("There already exists a file with the same name.\n
                The file was last updated on ",
                file.info(dmoz_file)$mtime,
                ".\n If you want to update the file, set overwrite to TRUE"))
  }

  tmp <- tempfile()
  curl_download(
    paste0("https://github.com/themains/rdomains/blob/master/data-raw/dmoz/",
           "dmoz_domain_category.zip?raw=TRUE"),
           tmp
           )
  unzip(tmp, exdir = outdir, overwrite = overwrite)
  unlink(tmp)

  cat("DMOZ Data saved to the following destination:", outdir, "\n")
}

#' Turn HTML into the text a classifier sees
#'
#' Mirrors the sibling Python package piedomains
#' (`src/piedomains/text_processor.py::clean_and_normalize_text`), whose contract as of
#' 0.12.0 is: **collapse runs of whitespace, trim, lowercase. Nothing else.**
#'
#' Two things it deliberately does *not* do, both of which were measured:
#'
#' * **Punctuation is kept.** Table pipes and layout dashes are 4.8% of tokens. Dropping
#'   them looked obviously right and measured worse -- held-out macro-F1 fell 0.7267 to
#'   0.7134 and Curlie agreement 0.543 to 0.523. Structural punctuation evidently says
#'   something about what kind of page this is.
#' * **No deduplication, sorting or stopword removal.** The pre-0.12.0 cleaner did all
#'   three; a transformer wants sentences, not a bag of sorted unique words.
#'
#' Do not "improve" this without retraining. The model is fitted to whichever form it was
#' shown, and a serving-side change silently degrades it -- which is a mistake this
#' project has already shipped once.
#'
#' @keywords internal
#' @noRd
NULL

#' Extract text, title, description and language from HTML
#'
#' @param html Raw HTML as a single string.
#'
#' @return A list with `text` (cleaned, lowercased), `title`, `description` and `lang`.
#'   Missing elements are `NA_character_`; `text` is `""` when nothing could be extracted.
#'
#' @export
#' @seealso [page_signals()], which uses this to decide whether a page is classifiable.
#' @examples
#' html <- "<html lang='en'><head><title>Example</title>
#'   <meta name='description' content='A demo page'>
#'   <script>ignored()</script></head>
#'   <body><p>Hello   World</p><style>p{}</style></body></html>"
#' html_text_content(html)
html_text_content <- function(html) {
  assert_character(html, len = 1, any.missing = FALSE)

  empty <- list(
    text = "", title = NA_character_,
    description = NA_character_, lang = NA_character_
  )
  if (!nzchar(str_trim(html))) {
    return(empty)
  }

  doc <- tryCatch(
    read_html(html),
    error = function(e) NULL
  )
  if (is.null(doc)) {
    return(empty)
  }

  # Remove what is not page text. Comments too: they carry build metadata and
  # occasionally whole alternate versions of the page.
  drop <- xml_find_all(
    doc, "//script | //style | //noscript | //template | //svg | //comment()"
  )
  if (length(drop)) {
    xml_remove(drop)
  }

  title <- first_text(doc, "//title")
  description <- first_attr(doc, "//meta[translate(@name,'DESCRIPTION','description')='description']", "content")
  lang <- first_attr(doc, "/html", "lang")

  body <- xml_find_first(doc, "//body")
  node <- if (inherits(body, "xml_missing")) doc else body

  # Collect text nodes individually and join with a space. Calling xml_text() on the body
  # concatenates across element boundaries, so "<h1>Example Domain</h1><p>This domain..."
  # came out as "example domainthis domain" -- two words fused into a token that appears
  # in no vocabulary. Python's extractor separates blocks, so this is also what parity
  # with the model's training input requires.
  pieces <- xml_text(xml_find_all(node, ".//text()"))
  raw_text <- paste(pieces[nzchar(str_trim(pieces))], collapse = " ")

  list(
    text = clean_page_text(raw_text),
    title = title,
    description = description,
    lang = lang
  )
}

#' Collapse whitespace, trim, lowercase -- and nothing else
#'
#' @param text raw extracted text
#' @return character(1)
#' @keywords internal
#' @noRd
clean_page_text <- function(text) {
  if (!length(text) || all(is.na(text))) {
    return("")
  }
  tolower(str_trim(str_replace_all(paste(text, collapse = " "), "\\s+", " ")))
}

#' First matching node's text, or NA
#' @keywords internal
#' @noRd
first_text <- function(doc, xpath) {
  node <- xml_find_first(doc, xpath)
  if (inherits(node, "xml_missing")) {
    return(NA_character_)
  }
  value <- str_trim(xml_text(node))
  if (nzchar(value)) value else NA_character_
}

#' First matching node's attribute, or NA
#' @keywords internal
#' @noRd
first_attr <- function(doc, xpath, attr) {
  node <- xml_find_first(doc, xpath)
  if (inherits(node, "xml_missing")) {
    return(NA_character_)
  }
  value <- xml_attr(node, attr)
  if (!is.na(value) && nzchar(str_trim(value))) str_trim(value) else NA_character_
}

test_that("script, style and comments are removed", {
  html <- "<html><body><p>real content here</p>
    <script>var x = 'javascript noise';</script>
    <style>.cls { color: red; }</style>
    <!-- a build comment --></body></html>"
  got <- html_text_content(html)$text
  expect_true(grepl("real content here", got, fixed = TRUE))
  expect_false(grepl("javascript noise", got, fixed = TRUE))
  expect_false(grepl("color", got, fixed = TRUE))
  expect_false(grepl("build comment", got, fixed = TRUE))
})

test_that("whitespace is collapsed and text lowercased, and nothing else", {
  # Mirrors piedomains' clean_and_normalize_text as of 0.12.0.
  got <- html_text_content("<html><body>Hello   \n\t  WORLD</body></html>")$text
  expect_equal(got, "hello world")
})

test_that("punctuation is kept, deliberately", {
  # Dropping table pipes and layout dashes looked obviously right and measured worse:
  # held-out macro-F1 0.7267 -> 0.7134. Do not "fix" this without retraining.
  got <- html_text_content("<html><body>Price: $10 | In stock -- yes!</body></html>")$text
  expect_true(grepl("|", got, fixed = TRUE))
  expect_true(grepl("--", got, fixed = TRUE))
  expect_true(grepl("$10", got, fixed = TRUE))
})

test_that("words are not deduplicated or sorted", {
  # The pre-0.12.0 cleaner did both. A transformer wants sentences.
  got <- html_text_content("<html><body>zebra apple zebra</body></html>")$text
  expect_equal(got, "zebra apple zebra")
})

test_that("title, description and lang are extracted", {
  html <- "<html lang='de'><head><title>Beispiel</title>
    <meta name='description' content='Eine Seite'></head><body>text</body></html>"
  got <- html_text_content(html)
  expect_equal(got$title, "Beispiel")
  expect_equal(got$description, "Eine Seite")
  expect_equal(got$lang, "de")
})

test_that("missing metadata comes back as NA, not empty string", {
  got <- html_text_content("<html><body>text</body></html>")
  expect_true(is.na(got$title))
  expect_true(is.na(got$description))
  expect_true(is.na(got$lang))
})

test_that("malformed and empty input does not error", {
  expect_equal(html_text_content("")$text, "")
  expect_equal(html_text_content("   ")$text, "")
  # Truncated mid-tag: xml2 is lenient, and we must not propagate a parse failure.
  expect_no_error(html_text_content("<html><body><p>unclosed"))
  expect_no_error(html_text_content("not html at all"))
})

test_that("a page with no body still yields its text", {
  got <- html_text_content("<html><head><title>T</title></head></html>")
  expect_type(got$text, "character")
})

#' Detect anti-bot interstitials, parking pages and empty placeholders
#'
#' Ported from the sibling Python package piedomains (`src/piedomains/blocking.py`). The
#' signatures were taken from pages that project actually received, not from documentation:
#' etsy.com, monster.com and reuters.com return a ~1470-byte DataDome interstitial whose
#' only visible text is the domain name; indeed.com returns Cloudflare's "Just a moment...";
#' mayoclinic.org returns a 296-byte Akamai "Access Denied ... Reference #".
#'
#' **Markers are tiered, because the presence of a vendor is not a block.** reddit, walmart,
#' tinyurl and quora all serve real pages while embedding reCAPTCHA, PerimeterX or a
#' Cloudflare Turnstile widget. Treating those as blocks silently discards good
#' classifications, so a weak marker only counts when the page also looks like an
#' interstitial. The tests carry those four as negative controls -- porting the tiers
#' without the counterexamples would port the bug back in.
#'
#' Everything here is a pure string heuristic over already-fetched HTML. No network.
#'
#' @keywords internal
#' @noRd
NULL

#' Unambiguous markers: these appear only on an actual challenge or denial page
#' @keywords internal
#' @noRd
RD_STRONG_MARKERS <- list(
  datadome   = c("captcha-delivery.com", "geo.captcha-delivery"),
  cloudflare = c("cf_chl_opt", "cf-browser-verification", "cf-please-wait"),
  imperva    = "incapsula incident id",
  akamai     = "reference #18."
)

#' Ambiguous markers: the vendor's script is present, but healthy pages embed these too
#' @keywords internal
#' @noRd
RD_WEAK_MARKERS <- list(
  cloudflare = "challenges.cloudflare.com",
  perimeterx = c("perimeterx", "px-captcha"),
  datadome   = "datadome",
  recaptcha  = c("g-recaptcha", "recaptcha/api.js")
)

#' Titles that are themselves the tell
#' @keywords internal
#' @noRd
RD_CHALLENGE_TITLES <- c(
  "just a moment...",
  "attention required! | cloudflare",
  "access denied",
  "security check",
  "are you a robot?",
  "verifying you are human"
)

#' HTTP statuses that usually mean "blocked", not "broken"
#' @keywords internal
#' @noRd
RD_BLOCK_STATUSES <- c(401L, 403L, 429L, 503L)

#' Deliberately narrow, and only applied to short pages: phrases like "permission to
#' access" appear in the legal copy of healthy sites (bankofamerica.com, 491 tokens,
#' matched an earlier looser pattern).
#' @keywords internal
#' @noRd
RD_DENIAL_TEXT <- paste0(
  "(you don'?t have permission to access|access denied",
  "|enable javascript and cookies to continue",
  "|verify you are (a )?human)"
)

#' Above this size a page is substantive enough that a weak marker is almost certainly an
#' embedded widget rather than an interstitial.
#' @keywords internal
#' @noRd
RD_INTERSTITIAL_MAX_BYTES <- 60000L

#' Registrars and parking services whose landing pages are unmistakable
#' @keywords internal
#' @noRd
RD_PARKING_SERVICES <- c(
  "dan.com", "sedo.com", "hugedomains", "afternic", "bodis.com",
  "parkingcrew", "wc_landing", "domainmarket", "undeveloped.com", "namesilo.com"
)

#' Phrases a parking page uses about itself. Held to a length limit because a real shop
#' legitimately says "for sale" while running to thousands of words.
#' @keywords internal
#' @noRd
RD_PARKING_PHRASES <- c(
  "is for sale", "buy this domain", "may be for sale", "domain for sale",
  "this domain is parked", "domain parking", "inquire about this domain",
  "domain name is for sale", "the domain you are looking for is for sale",
  # Found by sampling what the first version missed: orderwith.com, a parked page
  # labelled `drugs`, said "available for sale" rather than "is for sale".
  "available for sale", "available to purchase", "parked free of charge",
  "is available -- inquire"
)

#' A parked page is a placeholder. Above this it is a real site that mentions a sale.
#' @keywords internal
#' @noRd
RD_PARKED_MAX_WORDS <- 250L

#' What a web server emits when there is no site behind the domain
#' @keywords internal
#' @noRd
RD_SERVER_ARTIFACTS <- c(
  "index of /", "parent directory", "proudly served by", "litespeed web server",
  "welcome to nginx", "default web page", "it works!", "object not found!"
)

#' What a site says when it exists on paper but has no content yet.
#'
#' Multilingual because the corpus is. The Spanish and German entries are deliberately
#' truncated ("en construcci", not "en construccion" with an accent) -- that is how the
#' Python original handles the accent, and it also keeps this file pure ASCII, which
#' avoids an R CMD check note.
#' @keywords internal
#' @noRd
RD_UNAVAILABLE_PHRASES <- c(
  "coming soon", "under construction", "im aufbau", "en construcci", "en travaux",
  "account suspended", "site is offline", "seite ist offline", "maintenance page",
  "we will be back", "website expired", "bandwidth limit exceeded",
  "site temporarily unavailable", "page is under maintenance"
)

#' Far tighter than the parking limit, and it has to be. "Coming soon" and "under
#' construction" are things live sites say in passing: a 319-word breeder site saying
#' puppies are coming soon, an 877-word games portal. Sampled at 100 words the rule still
#' swept in a riding school; at 60 the first thirty hits were all genuine. Precision
#' matters more than recall, because a false positive relabels a real page.
#' @keywords internal
#' @noRd
RD_UNAVAILABLE_MAX_WORDS <- 60L

#' Extract a page title, lowercased
#' @param html raw response body
#' @return character(1), "" when absent
#' @keywords internal
#' @noRd
html_title <- function(html) {
  m <- str_match(html, regex("<title[^>]*>(.*?)</title>",
                                               ignore_case = TRUE, dotall = TRUE))
  if (is.na(m[1, 2])) "" else tolower(str_trim(m[1, 2]))
}

#' Does any marker appear in the body?
#' @param lowered lowercased html
#' @param markers named list of character vectors
#' @return the matching name, or NA
#' @keywords internal
#' @noRd
first_marker <- function(lowered, markers) {
  for (vendor in names(markers)) {
    if (any(vapply(markers[[vendor]], function(m) grepl(m, lowered, fixed = TRUE),
                   logical(1)))) {
      return(vendor)
    }
  }
  NA_character_
}

#' Corroborate a weak marker: does this page look like a challenge?
#' @param html raw response body
#' @param domain requested domain
#' @return logical
#' @keywords internal
#' @noRd
looks_like_interstitial <- function(html, domain = "") {
  title <- html_title(html)
  if (title %in% RD_CHALLENGE_TITLES) {
    return(TRUE)
  }
  if (nzchar(domain) && identical(title, tolower(str_trim(domain))) &&
      nchar(html) < 4000) {
    return(TRUE)
  }
  nchar(html) < RD_INTERSTITIAL_MAX_BYTES &&
    grepl(RD_DENIAL_TEXT, html, ignore.case = TRUE, perl = TRUE)
}

#' Is the extracted text too short to classify honestly?
#'
#' @param text Extracted page text.
#' @param min_tokens Floor, in whitespace-separated tokens.
#'
#' @return `TRUE` when the text falls below the floor.
#' @keywords internal
#' @noRd
is_thin <- function(text, min_tokens = 30L) {
  length(strsplit(str_trim(text), "\\s+")[[1]][nzchar(
    strsplit(str_trim(text), "\\s+")[[1]]
  )]) < min_tokens
}

#' Is this page a domain-parking placeholder rather than a site?
#'
#' 7.9% of the piedomains training corpus turned out to be parking pages, unevenly spread:
#' 42% of the `drugs` class, 23% of `webmail`, 18% of `downloads`, because expired domains
#' in those niches get parked. The model consequently learned that a "this domain is for
#' sale" template *means* drugs, which is why zappos.com came back as drugs.
#'
#' @param text Extracted page text.
#' @return `TRUE` when the page is a parking placeholder.
#' @keywords internal
#' @noRd
looks_parked <- function(text) {
  lowered <- tolower(text)
  if (any(vapply(RD_PARKING_SERVICES, function(s) grepl(s, lowered, fixed = TRUE),
                 logical(1)))) {
    return(TRUE)
  }
  # A phrase alone is not enough: length separates a placeholder from a shop.
  if (word_count(lowered) > RD_PARKED_MAX_WORDS) {
    return(FALSE)
  }
  any(vapply(RD_PARKING_PHRASES, function(p) grepl(p, lowered, fixed = TRUE), logical(1)))
}

#' Does the domain resolve but have no site behind it?
#'
#' Distinct from parked, which is specifically *for sale*. This is the other way a domain
#' serves bytes without being a website: an Apache autoindex, a registrar's "coming soon",
#' a suspended account.
#'
#' @param text Extracted page text.
#' @return `TRUE` when the page is a no-site placeholder.
#' @keywords internal
#' @noRd
looks_unavailable <- function(text) {
  lowered <- tolower(text)
  # Length first: these phrases are common inside real pages, and the guard is what
  # separates "we will be back after maintenance" from a shop mentioning a new line.
  if (word_count(lowered) > RD_UNAVAILABLE_MAX_WORDS) {
    return(FALSE)
  }
  any(vapply(RD_SERVER_ARTIFACTS, function(m) grepl(m, lowered, fixed = TRUE),
             logical(1))) ||
    any(vapply(RD_UNAVAILABLE_PHRASES, function(p) grepl(p, lowered, fixed = TRUE),
               logical(1)))
}

#' Count whitespace-separated words
#' @param x character(1)
#' @return integer
#' @keywords internal
#' @noRd
word_count <- function(x) {
  parts <- strsplit(str_trim(x), "\\s+")[[1]]
  sum(nzchar(parts))
}

#' What kind of page is this?
#'
#' Inspects already-fetched HTML for the three things that are not a classifiable site: an
#' anti-bot interstitial, a domain-parking placeholder, and a server's "nothing here"
#' page. No network access.
#'
#' `parked` and `unavailable` are **answers, not failures** -- they are facts about the
#' domain, plainly stated in the page, that a caller can act on. They are also free: no
#' classification service needs to be consulted.
#'
#' @param html Raw response body.
#' @param text Extracted page text. If `NULL`, derived from `html`.
#' @param domain The domain requested, used to spot a page whose only content is its own
#'   name.
#' @param status HTTP status code, if known.
#'
#' @return A one-row tibble: `page_state` (one of `"content"`, `"blocked"`, `"parked"`,
#'   `"unavailable"`, `"thin"`), `blocked`, `block_vendor`, `block_reason`, `parked`,
#'   `unavailable`, `thin`, `n_tokens`.
#'
#' @export
#' @seealso [fetch_error_codes()] for how these map onto run outcomes.
#' @examples
#' # A Cloudflare challenge, not a website
#' page_signals("<html><title>Just a moment...</title><body>cf_chl_opt</body></html>")
#'
#' # A parking page
#' page_signals("<html><body>This domain is for sale. Inquire now.</body></html>")
#'
#' # A real page that merely embeds reCAPTCHA is not blocked
#' page_signals(paste0("<html><title>Reddit</title><body>",
#'                     paste(rep("real discussion content", 200), collapse = " "),
#'                     "<script src='recaptcha/api.js'></script></body></html>"))
page_signals <- function(html, text = NULL, domain = "", status = NULL) {
  assert_character(html, len = 1, any.missing = FALSE)
  if (is.null(text)) {
    text <- html_text_content(html)$text
  }
  if (is.null(domain) || is.na(domain)) domain <- ""

  lowered <- tolower(html)

  vendor <- first_marker(lowered, RD_STRONG_MARKERS)
  blocked <- !is.na(vendor)
  reason <- if (blocked) "strong anti-bot marker" else NA_character_

  if (!blocked && !is.null(status) && !is.na(status) &&
      as.integer(status) %in% RD_BLOCK_STATUSES &&
      looks_like_interstitial(html, domain)) {
    blocked <- TRUE
    vendor <- "status"
    reason <- paste0("HTTP ", status, " with an interstitial body")
  }

  if (!blocked) {
    weak <- first_marker(lowered, RD_WEAK_MARKERS)
    if (!is.na(weak) && looks_like_interstitial(html, domain)) {
      blocked <- TRUE
      vendor <- weak
      reason <- "weak marker corroborated by interstitial shape"
    }
  }

  parked <- looks_parked(text)
  unavailable <- !parked && looks_unavailable(text)
  n_tokens <- word_count(text)
  thin <- n_tokens < 30L

  state <- if (blocked) {
    "blocked"
  } else if (parked) {
    "parked"
  } else if (unavailable) {
    "unavailable"
  } else if (thin) {
    "thin"
  } else {
    "content"
  }

  tibble(
    page_state = state,
    blocked = blocked,
    block_vendor = if (is.na(vendor)) NA_character_ else vendor,
    block_reason = reason,
    parked = parked,
    unavailable = unavailable,
    thin = thin,
    n_tokens = n_tokens
  )
}

# The negative controls are the point of this file.
#
# piedomains' blocking.py records the cases that cost it real classifications: reddit,
# walmart, tinyurl and quora all serve real pages while embedding reCAPTCHA, PerimeterX or
# a Cloudflare Turnstile widget; bankofamerica.com (491 tokens) matched an earlier looser
# denial pattern; a 319-word breeder site says puppies are "coming soon"; a real shop
# legitimately says "for sale".
#
# Porting the tiers without porting the counterexamples would port the bug back in.

long_page <- function(words, ...) {
  paste0(
    "<html><title>", ..1 %||% "A Real Site", "</title><body>",
    paste(rep("genuine page content about the subject", ceiling(words / 6)),
          collapse = " "),
    "</body></html>"
  )
}
`%||%` <- function(x, y) if (is.null(x)) y else x

test_that("strong anti-bot markers are detected", {
  expect_equal(page_signals("<html><body>captcha-delivery.com</body></html>")$block_vendor,
               "datadome")
  expect_equal(page_signals("<html><body>cf_chl_opt</body></html>")$block_vendor,
               "cloudflare")
  expect_equal(page_signals("<html><body>Incapsula incident ID 123</body></html>")$block_vendor,
               "imperva")
  expect_true(page_signals("<html><body>Reference #18.abc</body></html>")$blocked)
})

test_that("a challenge title is itself the tell", {
  s <- page_signals("<html><title>Just a moment...</title><body>challenges.cloudflare.com</body></html>")
  expect_true(s$blocked)
  expect_equal(s$page_state, "blocked")
})

test_that("a healthy page embedding reCAPTCHA is NOT blocked", {
  # reddit, walmart, tinyurl and quora all do this. Treating it as a block silently
  # discards good classifications.
  html <- paste0(
    "<html><title>Reddit - Dive into anything</title><body>",
    paste(rep("real discussion thread content here", 200), collapse = " "),
    "<script src='https://www.google.com/recaptcha/api.js'></script></body></html>"
  )
  s <- page_signals(html, domain = "reddit.com")
  expect_false(s$blocked)
  expect_equal(s$page_state, "content")
})

test_that("a long page mentioning access denial in legal copy is NOT blocked", {
  # bankofamerica.com, 491 tokens, matched an earlier looser pattern.
  html <- paste0(
    "<html><title>Bank of America</title><body>",
    paste(rep("banking services and account information", 150), collapse = " "),
    " You do not have permission to access certain areas without signing in. ",
    paste(rep("more legitimate banking copy", 150), collapse = " "),
    "<script src='recaptcha/api.js'></script></body></html>"
  )
  expect_false(page_signals(html, domain = "bankofamerica.com")$blocked)
})

test_that("parking pages are detected", {
  expect_true(page_signals("<html><body>This domain is for sale.</body></html>")$parked)
  expect_true(page_signals("<html><body>Buy this domain today</body></html>")$parked)
  # The one the first version missed: orderwith.com said "available for sale".
  expect_true(page_signals("<html><body>This name is available for sale</body></html>")$parked)
  # A parking service name is enough on its own, at any length.
  expect_true(page_signals("<html><body>Hosted by sedo.com</body></html>")$parked)
})

test_that("a real shop saying 'for sale' is NOT parked", {
  html <- paste0(
    "<html><title>Bike Shop</title><body>",
    "Our bikes are for sale in store and online. ",
    paste(rep("browse our range of road and mountain bicycles", 80), collapse = " "),
    "</body></html>"
  )
  s <- page_signals(html)
  expect_false(s$parked)
  expect_equal(s$page_state, "content")
})

test_that("empty placeholders are detected", {
  expect_true(page_signals("<html><body>Index of /</body></html>")$unavailable)
  expect_true(page_signals("<html><body>Welcome to nginx!</body></html>")$unavailable)
  expect_true(page_signals("<html><body>Account Suspended</body></html>")$unavailable)
  expect_true(page_signals("<html><body>Diese Seite ist noch im Aufbau</body></html>")$unavailable)
})

test_that("a real site saying 'coming soon' in passing is NOT unavailable", {
  # A 319-word breeder site saying puppies are coming soon; sampled at 100 words the
  # rule still swept in a riding school, which is why the floor is 60.
  html <- paste0(
    "<html><title>Kennel</title><body>",
    paste(rep("we breed labradors and show them at events", 60), collapse = " "),
    " New puppies coming soon. ",
    paste(rep("contact us to arrange a visit to the kennel", 40), collapse = " "),
    "</body></html>"
  )
  expect_false(page_signals(html)$unavailable)
})

test_that("thin pages are flagged but distinguished from blocked", {
  s <- page_signals("<html><body>Hello there friend</body></html>")
  expect_true(s$thin)
  expect_false(s$blocked)
  expect_equal(s$page_state, "thin")
})

test_that("page_state prefers blocked over parked over unavailable over thin", {
  both <- page_signals("<html><body>cf_chl_opt this domain is for sale</body></html>")
  expect_equal(both$page_state, "blocked")
})

test_that("page_signals returns one row with the documented columns", {
  s <- page_signals("<html><body>hello</body></html>")
  expect_equal(nrow(s), 1)
  expect_true(all(c("page_state", "blocked", "block_vendor", "parked",
                    "unavailable", "thin", "n_tokens") %in% names(s)))
})

#' rdomains: Classify Domains by their Content
#'
#' Want to know what kind of content is carried on a domain?
#' Get the results quickly using rdomains. The package provides access to virustotal
#' API, shalla, aws, OpenAI GPT models, Anthropic Claude models, 
#' and validated ML model based off shallalist data to predict content of a domain.
#'
#' To learn how to use rdomains, see this vignette: \url{../doc/rdomains.html}.
#'
#' @importFrom urltools suffix_extract
#' @importFrom Matrix Matrix spMatrix sparseVector
#' @importFrom jsonlite fromJSON toJSON
#' @importFrom stats setNames predict
#' @importFrom methods as
#' @importFrom glmnet predict.glmnet
#' @importFrom utils read.table URLencode untar unzip read.csv write.csv packageVersion
#' @importFrom curl curl_download nslookup
#' @importFrom httr content GET oauth_app oauth_signature add_headers POST status_code
#' @importFrom httr2 request req_user_agent req_timeout req_throttle req_retry
#'   req_error req_options req_perform resp_status resp_content_type resp_body_string
#'   resp_url url_parse resp_header
#' @importFrom xml2 read_xml as_list read_html xml_find_all xml_find_first xml_remove
#'   xml_text xml_attr url_absolute
#' @importFrom XML readHTMLTable
#' @importFrom virustotal set_key domain_report
#' @importFrom R.utils gunzip
#' @importFrom dplyr mutate bind_rows
#' @importFrom purrr map_chr map_df
#' @importFrom tibble tibble as_tibble tribble
#' @importFrom stringr str_trim str_remove str_detect str_match str_replace_all regex
#' @importFrom rlang abort warn inform .data
#' @importFrom cli cli_abort cli_warn cli_inform
#' @importFrom checkmate assert_character assert_file_exists assert_logical
#' @importFrom glue glue
#' @importFrom readr read_csv
#'
#' @author Gaurav Sood
"_PACKAGE"

#' Get Category from DMOZ
#'
#' Fetches category (or categories) of content hosted by a domain according to DMOZ.
#' The function checks if path to the DMOZ file is provided by the user.
#' If not, it looks for \code{dmoz_domain_cateory.csv} in the working directory. It also returns
#' results for prominent subdomains.
#'
#' DMOZ closed in March 2017 and the snapshot shipped here is from 2015, so these labels
#' describe the domain as it was a decade ago. The returned \code{source_last_published}
#' column carries that vintage, and a warning is issued once per session. See
#' \code{\link{source_vintage}}.
#'
#' @param domains vector of domain names
#' @param use_file path to the dmoz file, which can be downloaded using \code{\link{get_dmoz_data}}
#'
#' @return data.frame with the original list, the content category of the domain, and the
#'   vintage of the list that supplied it
#'
#' @export
#' @seealso \code{\link{source_vintage}} for the provenance of every category source
#' @importFrom dplyr coalesce
#' @examples \dontrun{
#' dmoz_cat(domains = "http://www.google.com")
#' dmoz_cat(domains = c("http://www.google.com", "http://plus.google.com"))
#' }

dmoz_cat <- function(domains = NULL, use_file = NULL) {

  validate_domains(domains)
  c_domains <- clean_domains(domains)
  c_domains_http <- str_remove(c_domains, "^http://")

  data_file <- validate_data_file(
    use_file,
    "dmoz_domain_category.csv",
    "get_dmoz_data"
  )

  dmoz <- read_csv(data_file, col_names = c("hostname", "category"), show_col_types = FALSE)

  warn_source_vintage("dmoz")

  tibble(
    domain_name = c_domains,
    dmoz_category = coalesce(
      dmoz$category[match(c_domains_http, dmoz$hostname)],
      dmoz$category[match(c_domains, dmoz$hostname)]
    ),
    source_last_published = SOURCE_VINTAGE$dmoz$last_published
  )
}

---
title: "Using rdomains"
author: "Gaurav Sood"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{Illustrating use of rdomains}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

### rdomains: Get the category of content hosted by a domain


#### Install and Load the package

The latest development version of the package will always be on GitHub. To install the package from GitHub and to load the installed package:

```{r, eval=FALSE, install}
#library(devtools)
install_github("themains/rdomains")
```

To install the package from CRAN, type in:

```{r, eval=FALSE, cran_install}
install.packages("rdomains")
```

Next, load the package:

```{r, eval=FALSE, load_pkg}
library(rdomains)
```

#### Shalla

To get category of the content from Shallalist (service discontinued - using archived data), first download the archived data using:

```{r, eval=FALSE, down_shalla}
get_shalla_data()
```

And then, get the category using:

```{r, eval=FALSE, shalla}
shalla_cat("http://www.google.com")
```

```
##   domain_name shalla_category
## 1  google.com   searchengines
```

#### DMOZ

To get category of the content from DMOZ, first download the archived parsed CSV file using:

```{r, eval=FALSE, down_dmoz}
get_dmoz_data()
```

And then, get the category using:

```{r, eval=FALSE, dmoz}
dmoz_cat("http://www.google.com")
```

#### ML

Probability that Domain Hosts Adult Content Based on features of Domain Name and Suffix alone:

```{r, eval=FALSE, ml}
adult_ml1_cat("http://www.google.com")
```

```
##   domain_name  category
## 1  google.com 0.3133728
```

#### VirusTotal

Start by getting the API key from [VirusTotal](https://www.virustotal.com/).

The package uses the VirusTotal API v3 for comprehensive domain analysis:

```{r, eval=FALSE, virustotal}
virustotal_cat("http://www.google.com")
```

#### OpenAI GPT Models

Get domain categorization using OpenAI's GPT models. You'll need an OpenAI API key:

```{r, eval=FALSE, openai}
# Set your API key
Sys.setenv("OPENAI_API_KEY", "your-api-key-here")

# Classify domains
openai_cat("google.com")
```

```
##   domain_name openai_category
## 1  google.com      technology
```

You can also specify custom categories:

```{r, eval=FALSE, openai_custom}
openai_cat(c("amazon.com", "github.com"), 
           categories = c("ecommerce", "technology", "social", "other"))
```

#### Anthropic Claude

Get domain categorization using Anthropic's Claude models. You'll need an Anthropic API key:

```{r, eval=FALSE, claude}
# Set your API key  
Sys.setenv("ANTHROPIC_API_KEY", "your-api-key-here")

# Classify domains
claude_cat("facebook.com")
```

```
##   domain_name claude_category
## 1 facebook.com          social
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/claude_cat.R
\name{claude_cat}
\alias{claude_cat}
\title{Get Category from Anthropic Claude}
\usage{
claude_cat(
  domains = NULL,
  api_key = NULL,
  categories = NULL,
  model = "claude-3-haiku-20240307",
  rate_limit = 0.5
)
}
\arguments{
\item{domains}{vector of domain names}

\item{api_key}{Anthropic API key. If not provided, looks for ANTHROPIC_API_KEY or CLAUDE_API_KEY environment variable}

\item{categories}{vector of categories to classify into. If NULL, uses default web categories}

\item{model}{Claude model to use (default: "claude-3-haiku-20240307" for cost efficiency)}

\item{rate_limit}{delay in seconds between API calls (default: 0.5)}
}
\value{
data.frame with original list and content category of the domain
}
\description{
Fetches category of content hosted by a domain using Anthropic's Claude API.
The function uses Claude models to classify domains into specified categories.
}
\examples{
\dontrun{
claude_cat("google.com")
claude_cat(c("google.com", "facebook.com"))
claude_cat("google.com", categories = c("search", "social", "ecommerce", "news", "other"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/openai_cat.R
\name{openai_cat}
\alias{openai_cat}
\title{Get Category from OpenAI}
\usage{
openai_cat(
  domains = NULL,
  api_key = NULL,
  categories = NULL,
  model = "gpt-4o-mini",
  rate_limit = 0.5
)
}
\arguments{
\item{domains}{vector of domain names}

\item{api_key}{OpenAI API key. If not provided, looks for OPENAI_API_KEY environment variable}

\item{categories}{vector of categories to classify into. If NULL, uses default web categories}

\item{model}{OpenAI model to use (default: "gpt-4o-mini" for cost efficiency)}

\item{rate_limit}{delay in seconds between API calls (default: 0.5)}
}
\value{
data.frame with original list and content category of the domain
}
\description{
Fetches category of content hosted by a domain using OpenAI's chat completion API.
The function uses GPT models to classify domains into specified categories.
}
\examples{
\dontrun{
openai_cat("google.com")
openai_cat(c("google.com", "facebook.com"))
openai_cat("google.com", categories = c("search", "social", "ecommerce", "news", "other"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/not_news.R
\name{not_news}
\alias{not_news}
\title{Classify News and Non-News Based on keywords in the URL}
\usage{
not_news(url_list = NULL)
}
\arguments{
\item{url_list}{vector of URLs}
}
\value{
data.frame with 3 columns: url, not_news, news
}
\description{
Based on a slightly amended version of the regular expression used to classify news, and non-news in:
``Exposure to ideologically diverse news and opinion on Facebook''
by Bakshy, Messing, and Adamic. Science. 2015.
}
\details{
Amendment: sport rather than sports

URL containing any of the following words is classified as soft news:
"sport|entertainment|arts|fashion|style|lifestyle|leisure|celeb|movie|music|gossip|food|travel|horoscope|weather|gadget"

URL containing any of following words is classified as hard news:
"politi|usnews|world|national|state|elect|vote|govern|campaign|war|polic|econ|unemploy|racis|energy|abortion|educa|healthcare|immigration"


Note that it is based on patterns existing in a small set of domains. See paper for details.
}
\examples{
\dontrun{
not_news("http://www.bbc.com/sport")
not_news(c("http://www.bbc.com/sport", "http://www.washingtontimes.com/news/politics/"))
}
}
\references{
\url{https://www.science.org/doi/10.1126/science.aaa1160}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/source_vintage.R
\name{source_vintage}
\alias{source_vintage}
\title{Report what each category source is and when it was last published}
\usage{
source_vintage(source = NULL)
}
\arguments{
\item{source}{Optional; one of \code{"shalla"}, \code{"dmoz"}, \code{"stevenblack"},
\code{"uni"}. If \code{NULL}, every source is returned.}
}
\value{
A tibble with one row per source: \code{key}, \code{source},
  \code{last_published}, \code{status}, \code{successor} and \code{note}.
}
\description{
Every lookup in this package answers from a static list, and those lists have very
different vintages. Two of them are no longer published at all. Use this to see which
answers can be trusted to describe a domain as it is today.
}
\examples{
source_vintage()
source_vintage("dmoz")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outcomes.R
\name{fetch_report}
\alias{fetch_report}
\title{Summarise the outcome of a fetch or classification run}
\usage{
fetch_report(x)
}
\arguments{
\item{x}{A tibble from [collect_content()] or [pie_cat()].}
}
\value{
A one-row tibble: `n`, `ok`, `failed`, `retryable`, and list-columns
  `by_stage`, `by_reason` and `missing` (the domains that produced nothing).
}
\description{
A pure function over the returned tibble, deliberately **not** a stored summary. A
snapshot goes stale the moment you filter the rows; recomputing means
`fetch_report(subset(x, retryable))` is always correct.
}
\examples{
rows <- data.frame(
  domain_name = c("a.com", "b.com", "c.com"),
  status = c("ok", "failed", "failed"),
  stage = c("infer", "fetch", "process"),
  error_code = c(NA, "timeout", "thin_content"),
  retryable = c(FALSE, TRUE, FALSE)
)
fetch_report(rows)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adult_ml1_cat.R
\name{adult_ml1_cat}
\alias{adult_ml1_cat}
\title{Probability that Domain Hosts Adult Content Based on features of Domain Name and Suffix alone.}
\usage{
adult_ml1_cat(domains = NULL)
}
\arguments{
\item{domains}{required; string; vector of domain names}
}
\value{
data.frame with original list and content category of the domains
}
\description{
Uses a validated ML model that uses keywords in the domain name
and suffix to predict probability that the domain hosts adult content. For
more information see \url{https://github.com/themains/keyword_porn}
}
\examples{
\dontrun{
adult_ml1_cat("http://www.google.com")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{glm_shalla}
\alias{glm_shalla}
\title{ML Model}
\format{
A list
}
\source{
ML model based on shallalist using keywords and domain suffixes,
}
\usage{
glm_shalla
}
\description{
ML Model
}
\author{
Gaurav Sood
}
\keyword{keywords}
\keyword{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_text.R
\name{html_text_content}
\alias{html_text_content}
\title{Extract text, title, description and language from HTML}
\usage{
html_text_content(html)
}
\arguments{
\item{html}{Raw HTML as a single string.}
}
\value{
A list with `text` (cleaned, lowercased), `title`, `description` and `lang`.
  Missing elements are `NA_character_`; `text` is `""` when nothing could be extracted.
}
\description{
Extract text, title, description and language from HTML
}
\examples{
html <- "<html lang='en'><head><title>Example</title>
  <meta name='description' content='A demo page'>
  <script>ignored()</script></head>
  <body><p>Hello   World</p><style>p{}</style></body></html>"
html_text_content(html)
}
\seealso{
[page_signals()], which uses this to decide whether a page is classifiable.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_dmoz_data.R
\name{get_dmoz_data}
\alias{get_dmoz_data}
\title{Get DMOZ Data}
\usage{
get_dmoz_data(outdir = ".", overwrite = FALSE)
}
\arguments{
\item{outdir}{Optional; folder to which you want to save the file; Default is same folder}

\item{overwrite}{Optional; default is FALSE. If TRUE, the file is overwritten.}
}
\description{
Downloads archived DMOZ (Open Directory Project) data. DMOZ was discontinued in March 2017.
This function downloads our preserved copy of the final DMOZ dataset. For more details, check:
\url{https://github.com/themains/rdomains/tree/master/data-raw/dmoz/}
}
\examples{
\dontrun{
get_dmoz_data()
}
}
\references{
\url{https://archive.org/details/dmoz-rdf-20150327}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shalla_cat.R
\name{shalla_cat}
\alias{shalla_cat}
\title{Get Category from Shallalist}
\usage{
shalla_cat(domains = NULL, use_file = NULL)
}
\arguments{
\item{domains}{vector of domain names}

\item{use_file}{path to the latest shallalist file downloaded using \code{\link{get_shalla_data}}}
}
\value{
data.frame with the original list, the content category of the domain, and the
  vintage of the list that supplied it
}
\description{
Fetches category of content hosted by a domain according to Shalla.
The function checks if path to the shalla file is provided by the user.
If not, it looks for \code{shalla_domain_category.csv} in the working directory.
}
\details{
Shallalist stopped publishing in 2022, so these labels describe the domain as it was
then or earlier, not as it is now. The returned \code{source_last_published} column
carries that vintage, and a warning is issued once per session. See
\code{\link{source_vintage}}.
}
\examples{
\dontrun{
shalla_cat(domains = "http://www.google.com")
}
}
\seealso{
\code{\link{source_vintage}} for the provenance of every category source
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_shalla_data.R
\name{get_shalla_data}
\alias{get_shalla_data}
\title{Get Shalla Data}
\usage{
get_shalla_data(outdir = "./", overwrite = FALSE)
}
\arguments{
\item{outdir}{Optional; folder to which you want to save the file; Default is same folder}

\item{overwrite}{Optional; default is FALSE. If TRUE, the file is overwritten.}
}
\description{
Shallalist service was discontinued in January 2022. This function downloads 
the last archived copy (from 1/14/22) that we have preserved on GitHub.
The original service at shallalist.de is no longer available.
Downloads, unzips and saves the final version of shallalist data. By default, saves shalla data  
as \code{shalla_domain_category.csv}.
}
\examples{
\dontrun{
get_shalla_data()
}
}
\references{
\url{https://web.archive.org/web/20210502020725/http://www.shallalist.de/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stevenblack_cat.R
\name{stevenblack_cat}
\alias{stevenblack_cat}
\title{Get Category from Steven Black's Host List}
\usage{
stevenblack_cat(domain = NULL, use_file = NULL)
}
\arguments{
\item{domain}{domain names as character vector}

\item{use_file}{path to a local Steven Black hosts file. If NULL, downloads from GitHub
once per session and reuses it}
}
\value{
data.frame with the original domain name, the category, and the date of the
  list that supplied it
}
\description{
Classifies domains based on Steven Black's unified host list which blocks
ads, malware, and tracking domains. The function checks if a domain appears
in the blocklist and categorizes it accordingly.
}
\details{
Steven Black's host list is a consolidated list from multiple sources including
adaway.org, mvps.org, malwaredomainlist.com, and someonewhocares.org.

Unlike the Shallalist and DMOZ lookups, this list is actively maintained, so the
returned \code{source_last_published} is the fetched file's own modification date
rather than a constant. See \code{\link{source_vintage}}.
}
\examples{
\dontrun{
stevenblack_cat("doubleclick.net")
stevenblack_cat(c("google.com", "googleadservices.com", "malware-example.com"))
}
}
\references{
\url{https://github.com/StevenBlack/hosts}
}
\seealso{
\code{\link{source_vintage}} for the provenance of every category source
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/page_signals.R
\name{page_signals}
\alias{page_signals}
\title{What kind of page is this?}
\usage{
page_signals(html, text = NULL, domain = "", status = NULL)
}
\arguments{
\item{html}{Raw response body.}

\item{text}{Extracted page text. If `NULL`, derived from `html`.}

\item{domain}{The domain requested, used to spot a page whose only content is its own
name.}

\item{status}{HTTP status code, if known.}
}
\value{
A one-row tibble: `page_state` (one of `"content"`, `"blocked"`, `"parked"`,
  `"unavailable"`, `"thin"`), `blocked`, `block_vendor`, `block_reason`, `parked`,
  `unavailable`, `thin`, `n_tokens`.
}
\description{
Inspects already-fetched HTML for the three things that are not a classifiable site: an
anti-bot interstitial, a domain-parking placeholder, and a server's "nothing here"
page. No network access.
}
\details{
`parked` and `unavailable` are **answers, not failures** -- they are facts about the
domain, plainly stated in the page, that a caller can act on. They are also free: no
classification service needs to be consulted.
}
\examples{
# A Cloudflare challenge, not a website
page_signals("<html><title>Just a moment...</title><body>cf_chl_opt</body></html>")

# A parking page
page_signals("<html><body>This domain is for sale. Inquire now.</body></html>")

# A real page that merely embeds reCAPTCHA is not blocked
page_signals(paste0("<html><title>Reddit</title><body>",
                    paste(rep("real discussion content", 200), collapse = " "),
                    "<script src='recaptcha/api.js'></script></body></html>"))
}
\seealso{
[fetch_error_codes()] for how these map onto run outcomes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_stevenblack_data.R
\name{get_stevenblack_data}
\alias{get_stevenblack_data}
\title{Get Steven Black's Host List Data}
\usage{
get_stevenblack_data(outdir = "./", variant = "base", overwrite = FALSE)
}
\arguments{
\item{outdir}{Optional; folder to which you want to save the file; Default is current directory}

\item{variant}{Optional; which variant to download. Options: "base", "porn", "social", "gambling", "all"}

\item{overwrite}{Optional; default is FALSE. If TRUE, the file is overwritten.}
}
\description{
Downloads the latest version of Steven Black's unified hosts file.
The hosts file contains domains known for serving ads, malware, and tracking.
}
\examples{
\dontrun{
get_stevenblack_data()
get_stevenblack_data(variant = "all")
}
}
\references{
\url{https://github.com/StevenBlack/hosts}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdomains.R
\docType{package}
\name{rdomains-package}
\alias{rdomains}
\alias{rdomains-package}
\title{rdomains: Classify Domains by their Content}
\description{
Want to know what kind of content is carried on a domain?
Get the results quickly using rdomains. The package provides access to virustotal
API, shalla, aws, OpenAI GPT models, Anthropic Claude models, 
and validated ML model based off shallalist data to predict content of a domain.
}
\details{
To learn how to use rdomains, see this vignette: \url{../doc/rdomains.html}.
}
\author{
Gaurav Sood
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collect_content.R
\name{collect_content}
\alias{collect_content}
\title{Fetch homepage HTML and text for domains}
\usage{
collect_content(
  domains = NULL,
  delay = 1,
  timeout = 10,
  max_bytes = 2 * 1024^2,
  obey_robots = TRUE,
  max_crawl_delay = 30,
  max_redirects = 5,
  user_agent = rdomains_user_agent()
)
}
\arguments{
\item{domains}{Character vector of domains or URLs.}

\item{delay}{Minimum seconds between requests to the same host. `Crawl-delay` overrides
this upward.}

\item{timeout}{Per-request timeout, seconds.}

\item{max_bytes}{Cap on the response body actually read.}

\item{obey_robots}{Whether to fetch and honour robots.txt. Turning this off is
discouraged and is your responsibility, not the package's.}

\item{max_crawl_delay}{Skip a host that asks for a longer delay than this rather than
sleeping on it.}

\item{max_redirects}{Maximum redirect hops to follow. Every hop is re-validated, so a
redirect cannot be used to reach an address the first check refused.}

\item{user_agent}{Override the identifying user-agent.}
}
\value{
A tibble with one row per input: `domain_name`, `status`, `stage`, `error_code`,
  `retryable`, `http_status`, `final_url`, `fetched_at`, `content_bytes`, `title`,
  `description`, `lang`, `text`, `n_tokens`, `page_state`, `block_vendor`,
  `robots_allowed`, `source_last_published`.
}
\description{
Every requested domain comes back, in order, whether or not it was reachable. Failures
carry a code from [fetch_error_codes()] rather than a bare `NA`.
}
\details{
The crawler identifies itself as `rdomains/<version>`, obeys `robots.txt` including
`Crawl-delay`, spaces requests to the same host, caps the response body, and refuses to
fetch hosts that resolve to private or link-local addresses.
}
\examples{
\dontrun{
res <- collect_content(c("example.com", "wikipedia.org"))
fetch_report(res)

# retry only what is worth retrying
again <- collect_content(res$domain_name[res$retryable])
}
}
\seealso{
[fetch_report()] to summarise the run, [page_signals()] for what the page
  states are, [source_vintage()] for how a live fetch compares with the static lists.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/virustotal_cat.R
\name{virustotal_cat}
\alias{virustotal_cat}
\title{Get Category from VirusTotal}
\usage{
virustotal_cat(domains = NULL, apikey = NULL)
}
\arguments{
\item{domains}{domain names as character vector}

\item{apikey}{virustotal API key}
}
\value{
data.frame with domain and VirusTotal analysis results
}
\description{
Returns category of content from multiple security vendors using the VirusTotal API v3.
The function retrieves domain analysis results including categories from various security
services. Not all services will have categories for all domains.
}
\details{
Get the API Access Key from \url{https://www.virustotal.com/}. Either pass the API Key to the function
or set the environmental variable: \code{VirustotalToken}. Environment variables persist within
a R session.
}
\examples{
\dontrun{
virustotal_cat("http://www.google.com")
virustotal_cat(c("google.com", "facebook.com"))
}
}
\references{
\url{https://docs.virustotal.com/reference/domains}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/uni_cat.R
\name{uni_cat}
\alias{uni_cat}
\title{Get Category from University Domain List}
\usage{
uni_cat(domains = NULL)
}
\arguments{
\item{domains}{vector of domain names}
}
\value{
data.frame with original list and all the other columns from the university json
}
\description{
Fetches university domain json from:
\url{https://raw.githubusercontent.com/Hipo/university-domains-list/master/world_universities_and_domains.json}
}
\examples{
\dontrun{
uni_cat(domains = "http://www.google.com")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dmoz_cat.R
\name{dmoz_cat}
\alias{dmoz_cat}
\title{Get Category from DMOZ}
\usage{
dmoz_cat(domains = NULL, use_file = NULL)
}
\arguments{
\item{domains}{vector of domain names}

\item{use_file}{path to the dmoz file, which can be downloaded using \code{\link{get_dmoz_data}}}
}
\value{
data.frame with the original list, the content category of the domain, and the
  vintage of the list that supplied it
}
\description{
Fetches category (or categories) of content hosted by a domain according to DMOZ.
The function checks if path to the DMOZ file is provided by the user.
If not, it looks for \code{dmoz_domain_cateory.csv} in the working directory. It also returns
results for prominent subdomains.
}
\details{
DMOZ closed in March 2017 and the snapshot shipped here is from 2015, so these labels
describe the domain as it was a decade ago. The returned \code{source_last_published}
column carries that vintage, and a warning is issued once per session. See
\code{\link{source_vintage}}.
}
\examples{
\dontrun{
dmoz_cat(domains = "http://www.google.com")
dmoz_cat(domains = c("http://www.google.com", "http://plus.google.com"))
}
}
\seealso{
\code{\link{source_vintage}} for the provenance of every category source
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outcomes.R
\name{fetch_error_codes}
\alias{fetch_error_codes}
\title{Reasons a domain can fail to get a category}
\usage{
fetch_error_codes()
}
\value{
A tibble with columns `code`, `stage`, `retryable` and `description`.
}
\description{
The closed set of `error_code` values that [collect_content()] and [pie_cat()] can
return, with the pipeline stage each belongs to and whether it is worth retrying.
}
\details{
These strings are shared verbatim with the Python package `piedomains`, so results from
the two can be aggregated together.
}
\examples{
fetch_error_codes()

# which failures are worth trying again?
subset(fetch_error_codes(), retryable)
}
\seealso{
[fetch_report()] to summarise an actual run, [source_vintage()] for the
  provenance of the static label sources.
}
